module intervals                                         ! Intervals in R

! Routines:
! 
! undefIntvlInt             Sets argument as undefined
! undefIntvlReal            Sets argument as undefined
! undefIntvlDbl             Sets argument as undefined
! undefIntvlQP              Sets argument as undefined
! isUndefIntvlInt           Returns: whether both lower and upper bound of an interval are undefined
! isUndefIntvlReal          Returns: whether both lower and upper bound of an interval are undefined
! isUndefIntvlDbl           Returns: whether both lower and upper bound of an interval are undefined
! isUndefIntvlQP            Returns: whether both lower and upper bound of an interval are undefined
! isDefIntvlInt             Returns: whether an interval is defined
! isDefIntvlReal            Returns: whether an interval is defined
! isDefIntvlDbl             Returns: whether an interval is defined
! isDefIntvlQP              Returns: whether an interval is defined
! setIfUndefIntvlInt        Sets variable to a given value if variable is undefined
! setIfUndefIntvlReal       Sets variable to a given value if variable is undefined
! setIfUndefIntvlDbl        Sets variable to a given value if variable is undefined
! setIfUndefIntvlQP         Sets variable to a given value if variable is undefined
! createCenteredIntvlDbl    Returns: interval centered around a given number and with a given width
! createCenteredIntvlReal   Returns: interval centered around a given number and with a given width
! widthInt                  Returns: interval width
! widthInt8                 Returns: interval width
! widthReal                 Returns: interval width
! widthDbl                  Returns: interval width
! widthQP                   Returns: interval width
! centralStretchIntvlDbl    Returns: interval derived from a given interval by central dilation
! centralStretchIntvlReal   Returns: interval derived from a given interval by central dilation
! intvlBoundsRatioDbl       Returns: ratio of upper and lower interval bound
! intvlCentreInt            Returns: interval centre
! intvlCentreInt8           Returns: interval centre
! intvlCentreReal           Returns: interval centre
! intvlCentreDbl            Returns: interval centre
! invBoundsInt              Returns: interval derived from a given interval by swapping its lower and upper bounds
! invBoundsInt8             Returns: interval derived from a given interval by swapping its lower and upper bounds
! invBoundsReal             Returns: interval derived from a given interval by swapping its lower and upper bounds
! invBoundsDbl              Returns: interval derived from a given interval by swapping its lower and upper bounds
! invBoundsQP               Returns: interval derived from a given interval by swapping its lower and upper bounds
! shiftIntvlReal            Returns: interval shifted w.r.t. a given interval by a given distance
! shiftIntvlDbl             Returns: interval shifted w.r.t. a given interval by a given distance
! stretchIntvlReal          Returns: interval derived from a given interval by multiplying both its bounds with a certain factor
! stretchIntvlDbl           Returns: interval derived from a given interval by multiplying both its bounds with a certain factor
! xtndIntvlReal             Returns: interval derived from a given interval by shifting both its bounds outward by a certain amount
! xtndIntvlDbl              Returns: interval derived from a given interval by shifting both its bounds outward by a certain amount
! putOntoIntvlInt           Returns: nearest number to a given number which lies within a certain interval
! putOntoIntvlInt8          Returns: nearest number to a given number which lies within a certain interval
! putOntoIntvlReal          Returns: nearest number to a given number which lies within a certain interval
! putOntoIntvlDbl           Returns: nearest number to a given number which lies within a certain interval
! isInIntvlInt              Returns: whether a given number lies within a certain interval
! isInIntvlInt8             Returns: whether a given number lies within a certain interval
! isInIntvlReal             Returns: whether a given number lies within a certain interval
! isInIntvlDbl              Returns: whether a given number lies within a certain interval
! putIntvlOntoIntvlReal     Returns: interval which results from shifting a given interval by the smallest possible distance
! intvlUnionReal            Returns: smallest contiguous superset of two given intervals
! intvlUnionDbl             Returns: smallest contiguous superset of two given intervals
! sectInt                   Returns: section of two intervals
! sectInt8                  Returns: section of two intervals
! sectReal                  Returns: section of two intervals
! sectDbl                   Returns: section of two intervals
! sectQP                    Returns: section of two intervals
! intvlIsInIntvlReal        Returns: whether a given interval lies in another given interval
! intvlIsInIntvlDbl         Returns: whether a given interval lies in another given interval

  use basicTypes
  use undefVals
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !=====================================================[ TYPE DEFINITIONS ]========================================================

  type tyIntvlInt
    integer :: &
     lBnd, &
     uBnd
    contains
      procedure :: invBounds => invBoundsInt
      procedure :: sect => sectInt
      procedure :: width => widthInt
  end type

  type tyIntvlInt8
    integer(i8b) :: &
     lBnd, &
     uBnd
    contains
      procedure :: invBounds => invBoundsInt8
      procedure :: sect => sectInt8
      procedure :: width => widthInt8
  end type

  type tyIntvlReal
    real :: &
     lBnd, &
     uBnd
    contains
      procedure :: invBounds => invBoundsReal
      procedure :: sect => sectReal
      procedure :: width => widthReal
  end type

  type tyIntvlDbl
    real(dp) :: &
     lBnd, &
     uBnd
    contains
      procedure :: invBounds => invBoundsDbl
      procedure :: sect => sectDbl
      procedure :: width => widthDbl
  end type

  type tyIntvlQP
    real(dp) :: &
     lBnd, &
     uBnd
    contains
      procedure :: invBounds => invBoundsQP
      procedure :: sect => sectQP
      procedure :: width => widthQP
  end type


  !========================================================[ CONSTANTS ]============================================================

  character(*),parameter :: &
   DESC_INTVL_L_BND="lower bound", &
   DESC_INTVL_U_BND="upper bound", &
   DESC_INTVL_RGE="span"


!––––––––––––––––––––––––––––––––––––––––––––––––––––––––––[ INTERFACES ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––


  ! UNDEFINED INTERVALS

  interface isUndef
    module procedure isUndefIntvlInt
    module procedure isUndefIntvlReal
    module procedure isUndefIntvlDbl
    module procedure isUndefIntvlQP
  end interface

  interface isDef
    module procedure isDefIntvlInt
    module procedure isDefIntvlReal
    module procedure isDefIntvlDbl
    module procedure isDefIntvlQP
  end interface

  interface undef
    module procedure undefIntvlInt
    module procedure undefIntvlReal
    module procedure undefIntvlDbl
    module procedure undefIntvlQP
  end interface

  interface setIfUndef
    module procedure setIfUndefIntvlInt
    module procedure setIfUndefIntvlReal
    module procedure setIfUndefIntvlDbl
    module procedure setIfUndefIntvlQP
  end interface


  ! CREATING INTERVALS

  interface createCenteredIntvl
    module procedure createCenteredIntvlReal
    module procedure createCenteredIntvlDbl
  end interface


  ! BASIC PROCEDURES

  interface centralStretchIntvl
    module procedure centralStretchIntvlReal
    module procedure centralStretchIntvlDbl
  end interface

  interface intvlBoundsRatio
    module procedure intvlBoundsRatioDbl
  end interface
  
  interface intvlCentre
    module procedure intvlCentreInt
    module procedure intvlCentreInt8
    module procedure intvlCentreReal
    module procedure intvlCentreDbl
  end interface

  interface shiftIntvl
    module procedure shiftIntvlReal
    module procedure shiftIntvlDbl
  end interface

  interface stretchIntvl
    module procedure stretchIntvlReal
    module procedure stretchIntvlDbl
  end interface

  interface xtndIntvl
    module procedure xtndIntvlReal
    module procedure xtndIntvlDbl
  end interface


  ! INTERVALS AND SCALARS

  interface putOntoIntvl
    module procedure putOntoIntvlInt
    module procedure putOntoIntvlInt8
    module procedure putOntoIntvlReal
    module procedure putOntoIntvlDbl
  end interface

  interface isInIntvl
    module procedure isInIntvlInt
    module procedure isInIntvlInt8
    module procedure isInIntvlReal
    module procedure isInIntvlDbl
  end interface


  ! COMBINING AND RELATING INTERVALS

  interface putIntvlOntoIntvl
    module procedure putIntvlOntoIntvlReal
  end interface

  interface intvlUnion
    module procedure intvlUnionReal
    module procedure intvlUnionDbl
  end interface

  interface intvlIsInIntvl
    module procedure intvlIsInIntvlReal
    module procedure intvlIsInIntvlDbl
  end interface


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


  !===================================================[ UNDEFINED INTERVALS ]=======================================================


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefIntvlInt(intvl)
!
!   sets argument as undefined
!
!   Arguments:
!     intvl                          var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlInt) :: &
     intvl
    intent(inout) :: &
     intvl

    intvl%lBnd=VAL_UNDEF_INT
    intvl%uBnd=VAL_UNDEF_INT

  end subroutine undefIntvlInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefIntvlReal(intvl)
!
!   sets argument as undefined
!
!   Arguments:
!     intvl                          var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlReal) :: &
     intvl
    intent(inout) :: &
     intvl

    intvl%lBnd=VAL_UNDEF_REAL
    intvl%uBnd=VAL_UNDEF_REAL

  end subroutine undefIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefIntvlDbl(intvl)
!
!   sets argument as undefined
!
!   Arguments:
!     intvl                          var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlDbl) :: &
     intvl
    intent(inout) :: &
     intvl

    intvl%lBnd=VAL_UNDEF_DBL
    intvl%uBnd=VAL_UNDEF_DBL

  end subroutine undefIntvlDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefIntvlQP(intvl)
!
!   sets argument as undefined
!
!   Arguments:
!     intvl                          var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlQP) :: &
     intvl
    intent(inout) :: &
     intvl

    intvl%lBnd=VAL_UNDEF_DBL
    intvl%uBnd=VAL_UNDEF_DBL

  end subroutine undefIntvlQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefIntvlInt(intvl)
!
!   Returns:                         whether both lower and upper bound of an interval are undefined
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isUndefIntvlInt
    type(tyIntvlInt) :: &
     intvl
    intent(in) :: &
     intvl

    isUndefIntvlInt=isUndef(intvl%lBnd) .OR. isUndef(intvl%uBnd)

  end function isUndefIntvlInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefIntvlReal(intvl)
!
!   Returns:                         whether both lower and upper bound of an interval are undefined
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isUndefIntvlReal
    type(tyIntvlReal) :: &
     intvl
    intent(in) :: &
     intvl

    isUndefIntvlReal=isUndef(intvl%lBnd) .OR. isUndef(intvl%uBnd)

  end function isUndefIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefIntvlDbl(intvl)
!
!   Returns:                         whether both lower and upper bound of an interval are undefined
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isUndefIntvlDbl
    type(tyIntvlDbl) :: &
     intvl
    intent(in) :: &
     intvl

    isUndefIntvlDbl=isUndef(intvl%lBnd) .OR. isUndef(intvl%uBnd)

  end function isUndefIntvlDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefIntvlQP(intvl)
!
!   Returns:                         whether both lower and upper bound of an interval are undefined
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isUndefIntvlQP
    type(tyIntvlQP) :: &
     intvl
    intent(in) :: &
     intvl

    isUndefIntvlQP=isUndef(intvl%lBnd) .OR. isUndef(intvl%uBnd)

  end function isUndefIntvlQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefIntvlInt(intvl)
!
!   Returns:                         whether an interval is defined
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isDefIntvlInt
    type(tyIntvlInt) :: &
     intvl
    intent(in) :: &
     intvl

    isDefIntvlInt=.NOT.isUndef(intvl)

  end function isDefIntvlInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefIntvlReal(intvl)
!
!   Returns:                         whether an interval is defined
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isDefIntvlReal
    type(tyIntvlReal) :: &
     intvl
    intent(in) :: &
     intvl

    isDefIntvlReal=.NOT.isUndef(intvl)

  end function isDefIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefIntvlDbl(intvl)
!
!   Returns:                         whether an interval is defined
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isDefIntvlDbl
    type(tyIntvlDbl) :: &
     intvl
    intent(in) :: &
     intvl

    isDefIntvlDbl=.NOT.isUndef(intvl)

  end function isDefIntvlDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefIntvlQP(intvl)
!
!   Returns:                         whether an interval is defined
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isDefIntvlQP
    type(tyIntvlQP) :: &
     intvl
    intent(in) :: &
     intvl

    isDefIntvlQP=.NOT.isUndef(intvl)

  end function isDefIntvlQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefIntvlInt(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!   intervals bounds treated independently
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlInt) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var%lBnd)) var%lBnd=val%lBnd
    if (isUndef(var%uBnd)) var%uBnd=val%uBnd

  end subroutine setIfUndefIntvlInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefIntvlReal(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!   intervals bounds treated independently
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlReal) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var%lBnd)) var%lBnd=val%lBnd
    if (isUndef(var%uBnd)) var%uBnd=val%uBnd

  end subroutine setIfUndefIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefIntvlDbl(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!   intervals bounds treated independently
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlDbl) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var%lBnd)) var%lBnd=val%lBnd
    if (isUndef(var%uBnd)) var%uBnd=val%uBnd

  end subroutine setIfUndefIntvlDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefIntvlQP(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!   intervals bounds treated independently
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlQP) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var%lBnd)) var%lBnd=val%lBnd
    if (isUndef(var%uBnd)) var%uBnd=val%uBnd

  end subroutine setIfUndefIntvlQP




  !===================================================[ CREATING INTERVALS ]========================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function createCenteredIntvlDbl(centre,width)
!
!   Returns:                         interval centered around a given number and with a given width
!
!   Arguments:
!     centre                         interval centre
!     width                          interval wirdt
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     centre, &
     width
    type(tyIntvlDbl) :: &
     createCenteredIntvlDbl
    intent(in) :: &
     centre, &
     width

    createCenteredIntvlDbl%lBnd=centre-width/2.
    createCenteredIntvlDbl%uBnd=centre+width/2.

  end function createCenteredIntvlDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function createCenteredIntvlReal(centre,width)
!
!   Returns:                         interval centered around a given number and with a given width
!
!   Arguments:
!     centre                         interval centre
!     width                          interval wirdt
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     centre, &
     width
    type(tyIntvlReal) :: &
     createCenteredIntvlReal
    intent(in) :: &
     centre, &
     width

    createCenteredIntvlReal%lBnd=centre-width/2.
    createCenteredIntvlReal%uBnd=centre+width/2.

  end function createCenteredIntvlReal




  !====================================================[ BASIC PROCEDURES ]=========================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function widthInt(intvl)
!
!   Returns:                         interval width
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     widthInt
    class(tyIntvlInt) :: &
     intvl

    widthInt=intvl%uBnd-intvl%lBnd

  end function widthInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function widthInt8(intvl)
!
!   Returns:                         interval width
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     widthInt8
    class(tyIntvlInt8) :: &
     intvl

    widthInt8=intvl%uBnd-intvl%lBnd

  end function widthInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function widthReal(intvl)
!
!   Returns:                         interval width
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     widthReal
    class(tyIntvlReal) :: &
     intvl

    widthReal=intvl%uBnd-intvl%lBnd

  end function widthReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function widthDbl(intvl)
!
!   Returns:                         interval width
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     widthDbl
    class(tyIntvlDbl) :: &
     intvl

    widthDbl=intvl%uBnd-intvl%lBnd

  end function widthDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function widthQP(intvl)
!
!   Returns:                         interval width
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(qp) :: &
     widthQP
    class(tyIntvlQP) :: &
     intvl

    widthQP=intvl%uBnd-intvl%lBnd

  end function widthQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function centralStretchIntvlDbl(intvl,fact)
!
!   Returns:                         interval derived from a given interval by central dilation
!
!   Arguments:
!     intvl                          original interval
!     fact                           scale factor
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     fact, &
     w, &
     fact2
    type(tyIntvlDbl) :: &
     intvl, &
     centralStretchIntvlDbl
    intent(in) :: &
     intvl, &
     fact

    w=intvl%uBnd-intvl%lBnd
    fact2=(fact-1.)/2.

    centralStretchIntvlDbl%lBnd=intvl%lBnd-fact2*w
    centralStretchIntvlDbl%uBnd=intvl%uBnd+fact2*w

  end function centralStretchIntvlDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function centralStretchIntvlReal(intvl,fact)
!
!   Returns:                         interval derived from a given interval by central dilation
!
!   Arguments:
!     intvl                          original interval
!     fact                           scale factor
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     fact, &
     w, &
     fact2
    type(tyIntvlReal) :: &
     intvl, &
     centralStretchIntvlReal
    intent(in) :: &
     intvl, &
     fact

    w=intvl%uBnd-intvl%lBnd
    fact2=(fact-1.)/2.

    centralStretchIntvlReal%lBnd=intvl%lBnd-fact2*w
    centralStretchIntvlReal%uBnd=intvl%uBnd+fact2*w

  end function centralStretchIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function intvlBoundsRatioDbl(intvl)
!
!   Returns:                         ratio of upper and lower interval bound
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     intvlBoundsRatioDbl
    type(tyIntvlDbl) :: &
     intvl
    intent(in) :: &
     intvl

    intvlBoundsRatioDbl=intvl%uBnd/intvl%lBnd

  end function intvlBoundsRatioDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function intvlCentreInt(intvl)
!
!   Returns:                         interval centre
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     intvlCentreInt
    type(tyIntvlInt) :: &
     intvl
    intent(in) :: &
     intvl
    
    intvlCentreInt=.5_dp*(intvl%lBnd+intvl%uBnd)
    
  end function intvlCentreInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function intvlCentreInt8(intvl)
!
!   Returns:                         interval centre
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     intvlCentreInt8
    type(tyIntvlInt8) :: &
     intvl
    intent(in) :: &
     intvl
    
    intvlCentreInt8=.5_dp*(intvl%lBnd+intvl%uBnd)
    
  end function intvlCentreInt8
    




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function intvlCentreReal(intvl)
!
!   Returns:                         interval centre
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     intvlCentreReal
    type(tyIntvlReal) :: &
     intvl
    intent(in) :: &
     intvl
    
    intvlCentreReal=.5*(intvl%lBnd+intvl%uBnd)
    
  end function intvlCentreReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function intvlCentreDbl(intvl)
!
!   Returns:                         interval centre
!
!   Arguments:
!     intvl                          interval in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     intvlCentreDbl
    type(tyIntvlDbl) :: &
     intvl
    intent(in) :: &
     intvl
    
    intvlCentreDbl=.5_dp*(intvl%lBnd+intvl%uBnd)
    
  end function intvlCentreDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function invBoundsInt(intvl)
!
!   Returns:                         interval derived from a given interval by swapping its lower and upper bounds
!
!   Arguments:
!     intvl                          original interval
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlInt) :: &
     invBoundsInt
    class(tyIntvlInt) :: &
     intvl
    intent(in) :: &
     intvl
    
    invBoundsInt%lBnd=intvl%uBnd
    invBoundsInt%uBnd=intvl%lBnd
    
  end function invBoundsInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function invBoundsInt8(intvl)
!
!   Returns:                         interval derived from a given interval by swapping its lower and upper bounds
!
!   Arguments:
!     intvl                          original interval
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlInt8) :: &
     invBoundsInt8
    class(tyIntvlInt8) :: &
     intvl
    intent(in) :: &
     intvl
    
    invBoundsInt8%lBnd=intvl%uBnd
    invBoundsInt8%uBnd=intvl%lBnd
    
  end function invBoundsInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function invBoundsReal(intvl)
!
!   Returns:                         interval derived from a given interval by swapping its lower and upper bounds
!
!   Arguments:
!     intvl                          original interval
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlReal) :: &
     invBoundsReal
    class(tyIntvlReal) :: &
     intvl
    intent(in) :: &
     intvl
    
    invBoundsReal%lBnd=intvl%uBnd
    invBoundsReal%uBnd=intvl%lBnd
    
  end function invBoundsReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function invBoundsDbl(intvl)
!
!   Returns:                         interval derived from a given interval by swapping its lower and upper bounds
!
!   Arguments:
!     intvl                          original interval
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlDbl) :: &
     invBoundsDbl
    class(tyIntvlDbl) :: &
     intvl
    intent(in) :: &
     intvl
    
    invBoundsDbl%lBnd=intvl%uBnd
    invBoundsDbl%uBnd=intvl%lBnd
    
  end function invBoundsDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function invBoundsQP(intvl)
!
!   Returns:                         interval derived from a given interval by swapping its lower and upper bounds
!
!   Arguments:
!     intvl                          original interval
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlQP) :: &
     invBoundsQP
    class(tyIntvlQP) :: &
     intvl
    intent(in) :: &
     intvl
    
    invBoundsQP%lBnd=intvl%uBnd
    invBoundsQP%uBnd=intvl%lBnd
    
  end function invBoundsQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function shiftIntvlReal(intvl,d)
!
!   Returns:                         interval shifted w.r.t. a given interval by a given distance
!
!   Arguments:
!     intvl                          original interval
!     d                              distance to shift
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     d
    type(tyIntvlReal) :: &
     shiftIntvlReal, &
     intvl
    intent(in) :: &
     intvl, &
     d

    shiftIntvlReal%lBnd=intvl%lBnd+d
    shiftIntvlReal%uBnd=intvl%uBnd+d

  end function shiftIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function shiftIntvlDbl(intvl,d)
!
!   Returns:                         interval shifted w.r.t. a given interval by a given distance
!
!   Arguments:
!     intvl                          original interval
!     d                              distance to shift
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     d
    type(tyIntvlDbl) :: &
     shiftIntvlDbl, &
     intvl
    intent(in) :: &
     intvl, &
     d

    shiftIntvlDbl%lBnd=intvl%lBnd+d
    shiftIntvlDbl%uBnd=intvl%uBnd+d

  end function shiftIntvlDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function stretchIntvlReal(intvl,fact)
!
!   Returns:                         interval derived from a given interval by multiplying both its bounds with a certain factor
!
!   Arguments:
!     intvl                          original interval
!     fact                           scale factor
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     fact
    type(tyIntvlReal) :: &
     intvl, &
     stretchIntvlReal
    intent(in) :: &
     intvl, &
     fact

    stretchIntvlReal%lBnd=intvl%lBnd*fact
    stretchIntvlReal%uBnd=intvl%uBnd*fact

  end function stretchIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function stretchIntvlDbl(intvl,fact)
!
!   Returns:                         interval derived from a given interval by multiplying both its bounds with a certain factor
!
!   Arguments:
!     intvl                          original interval
!     fact                           scale factor
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     fact
    type(tyIntvlDbl) :: &
     intvl, &
     stretchIntvlDbl
    intent(in) :: &
     intvl, &
     fact

    stretchIntvlDbl%lBnd=intvl%lBnd*fact
    stretchIntvlDbl%uBnd=intvl%uBnd*fact

  end function stretchIntvlDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function xtndIntvlReal(intvl,x)
!
!   Returns:                         interval derived from a given interval by shifting both its bounds outward by a certain amount
!
!   Arguments:
!     intvl                          original interval
!     x                              absolute value of distance to shift
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     x
    type(tyIntvlReal) :: &
     xtndIntvlReal, &
     intvl
    intent(in) :: &
     intvl, &
     x

    xtndIntvlReal%lBnd=intvl%lBnd-x
    xtndIntvlReal%uBnd=intvl%uBnd+x

  end function xtndIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function xtndIntvlDbl(intvl,x)
!
!   Returns:                         interval derived from a given interval by shifting both its bounds outward by a certain amount
!
!   Arguments:
!     intvl                          original interval
!     x                              absolute value of distance to shift
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     x
    type(tyIntvlDbl) :: &
     xtndIntvlDbl, &
     intvl
    intent(in) :: &
     intvl, &
     x

    xtndIntvlDbl%lBnd=intvl%lBnd-x
    xtndIntvlDbl%uBnd=intvl%uBnd+x

  end function xtndIntvlDbl




  !=================================================[ INTERVALS AND SCALARS ]=======================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function putOntoIntvlInt(intvl,nr)
!
!   Returns:                         nearest number to a given number which lies within a certain interval
!
!   Arguments:
!     intvl                          interval in question
!     nr                             original number
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     nr, &
     putOntoIntvlInt
    type(tyIntvlInt) :: &
     intvl
    intent(in) :: &
     intvl, &
     nr

    putOntoIntvlInt=merge(nr,merge(intvl%lBnd,intvl%uBnd,nr < intvl%lBnd),isInIntvl(intvl,nr))

  end function putOntoIntvlInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function putOntoIntvlInt8(intvl,nr)
!
!   Returns:                         nearest number to a given number which lies within a certain interval
!
!   Arguments:
!     intvl                          interval in question
!     nr                             original number
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     nr, &
     putOntoIntvlInt8
    type(tyIntvlInt8) :: &
     intvl
    intent(in) :: &
     intvl, &
     nr

    putOntoIntvlInt8=merge(nr,merge(intvl%lBnd,intvl%uBnd,nr < intvl%lBnd),isInIntvl(intvl,nr))

  end function putOntoIntvlInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function putOntoIntvlReal(intvl,nr)
!
!   Returns:                         nearest number to a given number which lies within a certain interval
!
!   Arguments:
!     intvl                          interval in question
!     nr                             original number
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     nr, &
     putOntoIntvlReal
    type(tyIntvlReal) :: &
     intvl
    intent(in) :: &
     intvl, &
     nr

    putOntoIntvlReal=merge(nr,merge(intvl%lBnd,intvl%uBnd,nr < intvl%lBnd),isInIntvl(intvl,nr))

  end function putOntoIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function putOntoIntvlDbl(intvl,nr)
!
!   Returns:                         nearest number to a given number which lies within a certain interval
!
!   Arguments:
!     intvl                          interval in question
!     nr                             original number
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     nr, &
     putOntoIntvlDbl
    type(tyIntvlDbl) :: &
     intvl
    intent(in) :: &
     intvl, &
     nr

    putOntoIntvlDbl=merge(nr,merge(intvl%lBnd,intvl%uBnd,nr < intvl%lBnd),isInIntvl(intvl,nr))

  end function putOntoIntvlDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isInIntvlInt(intvl,nr)
!
!   Returns:                         whether a given number lies within a certain interval
!
!   Arguments:
!     intvl                          interval in question
!     nr                             number in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     nr
    logical :: &
     isInIntvlInt
    type(tyIntvlInt) :: &
     intvl
    intent(in) :: &
     intvl, &
     nr

    isInIntvlInt=intvl%lBnd <= nr .AND. nr <= intvl%uBnd

  end function isInIntvlInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isInIntvlInt8(intvl,nr)
!
!   Returns:                         whether a given number lies within a certain interval
!
!   Arguments:
!     intvl                          interval in question
!     nr                             number in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     nr
    logical :: &
     isInIntvlInt8
    type(tyIntvlInt8) :: &
     intvl
    intent(in) :: &
     intvl, &
     nr

    isInIntvlInt8=intvl%lBnd <= nr .AND. nr <= intvl%uBnd

  end function isInIntvlInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isInIntvlReal(intvl,nr,openL,openR)
!
!   Returns:                         whether a given number lies within a certain interval
!
!   Arguments:
!     intvl                          interval in question
!     nr                             number in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     nr
    logical :: &
     isInIntvlReal, &
     openL, &
     openR, &
     okL, &
     okR
    type(tyIntvlReal) :: &
     intvl
    optional :: &
     openL, &
     openR
    intent(in) :: &
     intvl, &
     nr, &
     openL, &
     openR

    if (present(openL) .AND. openL) then
      okL=intvl%lBnd < nr
    else
      okL=intvl%lBnd <= nr
    end if

    if (present(openR) .AND. openR) then
      okR=intvl%uBnd > nr
    else
      okR=intvl%uBnd >= nr
    end if

    isInIntvlReal=okL .AND. okR

  end function isInIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isInIntvlDbl(intvl,nr,openL,openR)
!
!   Returns:                         whether a given number lies within a certain interval
!
!   Arguments:
!     intvl                          interval in question
!     nr                             number in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     nr
    logical :: &
     isInIntvlDbl, &
     openL, &
     openR, &
     okL, &
     okR
    type(tyIntvlDbl) :: &
     intvl
    optional :: &
     openL, &
     openR
    intent(in) :: &
     intvl, &
     nr, &
     openL, &
     openR

    if (present(openL) .AND. openL) then
      okL=intvl%lBnd < nr
    else
      okL=intvl%lBnd <= nr
    end if

    if (present(openR) .AND. openR) then
      okR=intvl%uBnd > nr
    else
      okR=intvl%uBnd >= nr
    end if

    isInIntvlDbl=okL .AND. okR

  end function isInIntvlDbl




  !============================================[ COMBINING AND RELATING INTERVALS ]=================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function putIntvlOntoIntvlReal(intvl1,intvl2)
!
!   Returns:                         interval which results from shifting a given interval by the smallest possible distance
!                                    s.t. it lies within another given interval
!
!   Arguments:
!     intvl1                         original interval
!     intvl2                         interval to contain the result
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     d
    type(tyIntvlReal) :: &
     putIntvlOntoIntvlReal, &
     intvl1, &
     intvl2
    intent(in) :: &
     intvl1, &
     intvl2

    if (intvl1%width() > intvl2%width() .OR. intvlIsInIntvl(intvl1,intvl2)) then ! putting intvl1 into intvl2 impossible / unnecessary
      putIntvlOntoIntvlReal=intvl1
    else
      if (intvl1%lBnd < intvl2%lBnd) then
        d=intvl2%lBnd-intvl1%lBnd
      else if (intvl1%uBnd > intvl2%uBnd) then
        d=intvl1%uBnd-intvl2%uBnd
      end if
      putIntvlOntoIntvlReal=shiftIntvl(intvl1,d)
    end if

  end function putIntvlOntoIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function intvlUnionReal(intvl1,intvl2)
!
!   Returns:                         smallest contiguous superset of two given intervals
!
!   Arguments:
!     intvl1                         interval
!     intvl2                         interval
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlReal) :: &
     intvl1, &
     intvl2, &
     intvlUnionReal
    intent(in) :: &
     intvl1, &
     intvl2

    intvlUnionReal=tyIntvlReal(min(intvl1%lBnd,intvl2%lBnd),max(intvl1%uBnd,intvl2%uBnd))

  end function intvlUnionReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function intvlUnionDbl(intvl1,intvl2)
!
!   Returns:                         smallest contiguous superset of two given intervals
!
!   Arguments:
!     intvl1                         interval
!     intvl2                         interval
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlDbl) :: &
     intvl1, &
     intvl2, &
     intvlUnionDbl
    intent(in) :: &
     intvl1, &
     intvl2

    intvlUnionDbl=tyIntvlDbl(min(intvl1%lBnd,intvl2%lBnd),max(intvl1%uBnd,intvl2%uBnd))

  end function intvlUnionDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function sectInt(intvl1,intvl2)
!
!   Returns:                         section of two intervals
!
!   Arguments:
!     intvl1                         interval
!     intvl2                         interval
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlInt) :: &
     sectInt, &
     intvl2
    class(tyIntvlInt) :: &
     intvl1
    intent(in) :: &
     intvl2

    sectInt=tyIntvlInt(max(intvl1%lBnd,intvl2%lBnd),min(intvl1%uBnd,intvl2%uBnd))
    if (sectInt%width() < 0) call undef(sectInt)

  end function sectInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function sectInt8(intvl1,intvl2)
!
!   Returns:                         section of two intervals
!
!   Arguments:
!     intvl1                         interval
!     intvl2                         interval
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlInt) :: &
     sectInt8, &
     intvl2
    class(tyIntvlInt8) :: &
     intvl1
    intent(in) :: &
     intvl2

    sectInt8=tyIntvlInt(max(intvl1%lBnd,intvl2%lBnd),min(intvl1%uBnd,intvl2%uBnd))
    if (sectInt8%width() < 0) call undef(sectInt8)

  end function sectInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function sectReal(intvl1,intvl2)
!
!   Returns:                         section of two intervals
!
!   Arguments:
!     intvl1                         interval
!     intvl2                         interval
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlReal) :: &
     sectReal, &
     intvl2
    class(tyIntvlReal) :: &
     intvl1
    intent(in) :: &
     intvl2

    sectReal=tyIntvlReal(max(intvl1%lBnd,intvl2%lBnd),min(intvl1%uBnd,intvl2%uBnd))
    if (sectReal%width() < 0) call undef(sectReal)

  end function sectReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function sectDbl(intvl1,intvl2)
!
!   Returns:                         section of two intervals
!
!   Arguments:
!     intvl1                         interval
!     intvl2                         interval
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlDbl) :: &
     sectDbl, &
     intvl2
    class(tyIntvlDbl) :: &
     intvl1
    intent(in) :: &
     intvl2

    sectDbl=tyIntvlDbl(max(intvl1%lBnd,intvl2%lBnd),min(intvl1%uBnd,intvl2%uBnd))
    if (sectDbl%width() < 0) call undef(sectDbl)

  end function sectDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function sectQP(intvl1,intvl2)
!
!   Returns:                         section of two intervals
!
!   Arguments:
!     intvl1                         interval
!     intvl2                         interval
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyIntvlQP) :: &
     sectQP, &
     intvl2
    class(tyIntvlQP) :: &
     intvl1
    intent(in) :: &
     intvl2

    sectQP=tyIntvlQP(max(intvl1%lBnd,intvl2%lBnd),min(intvl1%uBnd,intvl2%uBnd))
    if (sectQP%width() < 0) call undef(sectQP)

  end function sectQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function intvlIsInIntvlReal(intvl1,intvl2)
!
!   Returns:                         whether a given interval lies in another given interval
!
!   Arguments:
!     intvl1                         interval whose being contained is in question
!     intvl2                         interval whose being a superset is in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     intvlIsInIntvlReal
    type(tyIntvlReal) :: &
     intvl1, &
     intvl2
    intent(in) :: &
     intvl1, &
     intvl2

    intvlIsInIntvlReal=intvl1%lBnd >= intvl2%lBnd .AND. intvl1%uBnd <= intvl2%uBnd

  end function intvlIsInIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function intvlIsInIntvlDbl(intvl1,intvl2)
!
!   Returns:                         whether a given interval lies in another given interval
!
!   Arguments:
!     intvl1                         interval whose being contained is in question
!     intvl2                         interval whose being a superset is in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     intvlIsInIntvlDbl
    type(tyIntvlDbl) :: &
     intvl1, &
     intvl2
    intent(in) :: &
     intvl1, &
     intvl2

    intvlIsInIntvlDbl=intvl1%lBnd >= intvl2%lBnd .AND. intvl1%uBnd <= intvl2%uBnd

  end function intvlIsInIntvlDbl




end module
