module io                                            ! (Formatted) in-/output

! Routines:
! 
! indentPrefx            Returns: indentation prefix for standard output and log file
! startSubsect           Starts new subsection in standard output and/or log file
! endSubsect             Ends current subsection in standard output and/or log file
! fstStr                 Returns: first string after indentation prefix for standard output or log file
! printInt               Prints a quantity, obeying section level
! printInt2              Prints a quantity, obeying section level
! printInt8              Prints a quantity, obeying section level
! printReal              Prints a quantity, obeying section level
! printDbl               Prints a quantity, obeying section level
! printQP                Prints a quantity, obeying section level
! printBool              Prints a quantity, obeying section level
! printStr               Prints a string, obeying section level
! printIntvlInt          Prints a quantity, obeying section level
! printIntvlReal         Prints a quantity, obeying section level
! printIntvlDbl          Prints a quantity, obeying section level
! printIntvlQP           Prints a quantity, obeying section level
! printUncQtyDbl         Prints a quantity, obeying section level
! printUncQtyQP          Prints a quantity, obeying section level
! printMsg               Prints a message, obeying section level
! warn                   Prints a warning message, obeying section level
! printHostProps         Prints properties of host computer
! printPrgrs4            Prints information about the progress of some task
! printPrgrs8            Prints information about the progress of some task
! filNmPrefx             Returns: generic file-name prefix
! fmtFilNm               Returns: string formatted for use in a file name
! readCSVArrDbl          Returns: whether a given CSV file has no data left, after reading values from it
! writeCSVIntReal        Writes values to CSV file
! writeCSVIntDbl         Writes values to CSV file
! writeCSVInt8Dbl        Writes values to CSV file
! writeCSVInt8QP         Writes values to CSV file
! writeCSVRealReal       Writes values to CSV file
! writeCSVDblDbl         Writes values to CSV file
! writeArrCSVDbl         Writes values to CSV file
! writeArrCSVStr         Writes values to CSV file
! readIntQty             Reads a quantity from standard input after printing its description to standard output
! readRealQty            Reads a quantity from standard input after printing its description to standard output
! readDblQty             Reads a quantity from standard input after printing its description to standard output
! readBoolQty            Reads a quantity from standard input after printing its description to standard output
! readStrQty             Reads a quantity from standard input after printing its description to standard output

  use basicIO
  use uncert
  use env
  use time
  use maths
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !=====================================================[ TYPE DEFINITIONS ]========================================================

  type tyRuntmCtrlFilNm
    character(STR_LEN_STD) :: finish
  end type


  !========================================================[ CONSTANTS ]============================================================

  integer,parameter :: &
   CONS_TAB_STOP_STD=95, &
   CONS_TAB_STOP_OPT_STD=40, &
   CONS_VAL_LEN_MIN=20, &
   CONS_OPT_LEN_MIN=30, &
   OPT_DESC_SPC=2, &                 ! min. spacing between CL option and its description when printing help
   RULER_N_SPACES=2, &
   CONS_TAB_WIDTH=2, &
   CONS_LINE_LEN_MIN=25, &
   CONS_LINE_LEN_STD=125, &
   DISP_TBL_N_COLS=10, &
   DISP_TBL_COL_W(DISP_TBL_N_COLS)=[9,9,7,12,8,6,7,11,7,8], &
   DISP_TBL_COL_W_MAX=28, &          ! must be equal to the maximum width of any (combined) column
   DISP_TBL_PADD=0, &
   N_FILS_OUT_MAX=1000, &
   UN_CONF=10, &
   UN_DT=11, &                       ! for all data files
   UN_TMP=12, &
   UN_SAV_SAMP=18, &
   UN_RD_SAMP=19, &
   UN_SAV_CAND_CSV=200, &            ! CAUTION: don't use units >= 200 for other purposes (reserved for candidates)
   UN_RND=95, &
   UN_PLT_CSV=98, &
   UN_GNUPLT=99, &
   UN_TEX=97

  real,parameter :: &
   DISP_UPD_INTVL_DUR_S=1., &        ! # seconds between two display updates (only relevant if > MULTIPT_BLK_DUR_S)
   DISP_UPD_ETR_SELDOM_S=30., &
   DISP_UPD_ETR_LONG_S=10.*60.

  logical,parameter :: &
   DISP_TBL_NO_V_BAR=.TRUE.

  character(*),parameter :: &
   COMMENT_PREFX="#", &
   FIL_NM_ILLEGAL_CHARS="'?*/\ """, &
   OUT_DIR_STD="../out/", &          ! CAUTION: when changing this, also change corresponding help text
   SUBDIR_CSV_STD="csv/", &          ! CAUTION: when changing this, also change corresponding help text
   FIL_EXT_CSV=".csv", &
   FIL_EXT_SAV_SAMP=".base", &
   NEW_LIN=new_line("a"), &
   DISP_TBL_HDR1(DISP_TBL_N_COLS)=[character(15) :: "Elapsed","Remain","Speed","Cold-chain","Final samples", &
   repeat(VAL_UNDEF_CHR,15),"Burn-","Max. PSR",repeat(VAL_UNDEF_CHR,15),repeat(VAL_UNDEF_CHR,15)], &
   DISP_TBL_HDR2(DISP_TBL_N_COLS)=[character(9) :: "[h:m:s]","[h:m:s]","[1/s]","acc. rate","total","eff.","in","value","par.", &
   "recalc"], &
   CSV_SEP=char(9), &
   STD_FIL_NM_CONF="base.conf", &
   CMD_PREFX="$", &
   CMD_PREFX_HEADER=COMMENT_PREFX//CMD_PREFX, &
   DATA_DIR_STD="", &
   FIL_NM_RND_SEED=".BASE.randomSeed"


  !========================================================[ VARIABLES ]============================================================

  integer :: &
   sectDepth(0:1), &
   consLineLen, &
   consTabStop, &
   consTabStopOpt

  logical :: &
   filOutputPlt

  character(MAX_LEN_PATH) :: &
   outDir, &
   outDirCSV, &
   filNmLog, &
   curFilNmPrefx, &
   runIdent, &
   dataDir

  type(tyRuntmCtrlFilNm) :: &
   RUNTM_CTRL_FIL_NM=tyRuntmCtrlFilNm(".BASE.finish")


!––––––––––––––––––––––––––––––––––––––––––––––––––––––––––[ INTERFACES ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––


  !===================================================[ CONSOLE AND LOG FILE ]======================================================

  interface printQty
    module procedure printInt
    module procedure printInt2
    module procedure printInt8
    module procedure printReal
    module procedure printDbl
    module procedure printQP
    module procedure printBool
    module procedure printStr
    module procedure printIntvlInt
    module procedure printIntvlReal
    module procedure printIntvlDbl
    module procedure printIntvlQP
    module procedure printUncQtyDbl
    module procedure printUncQtyQP
  end interface

  interface printPrgrs
    module procedure printPrgrs4
    module procedure printPrgrs8
  end interface


  !=======================================================[ OTHER FILES ]===========================================================

  interface readCSVArr
    module procedure readCSVArrDbl
  end interface

  interface writeCSV
    module procedure writeCSVIntReal
    module procedure writeCSVIntDbl
    module procedure writeCSVInt8Dbl
    module procedure writeCSVInt8QP
    module procedure writeCSVRealReal
    module procedure writeCSVDblDbl
  end interface

  interface writeArrCSV
    module procedure writeArrCSVDbl
    module procedure writeArrCSVStr
  end interface


  !======================================================[ INTERACTIVITY ]==========================================================

  interface readQty
    module procedure readIntQty
    module procedure readRealQty
    module procedure readDblQty
    module procedure readBoolQty
    module procedure readStrQty
  end interface


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


  !=======================================================[ SECTIONING ]============================================================


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function indentPrefx(i,simp,spc)
!
!   Returns:                         indentation prefix for standard output and log file
!
!   Arguments:
!     i                              output-unit index
!     simp                           whether simple output (omitting the "+" on the left) is requested
!     spc                            whether prefix should only consist of spaces

!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    logical :: &
     simp, &
     spc
    character(sectDepth(i)*CONS_TAB_WIDTH) :: &
     indentPrefx
    optional :: &
     simp, &
     spc
    intent(in) :: &
     i, &
     simp, &
     spc

    if (sectDepth(i) == 0) then
      indentPrefx=""
    else if (present(simp) .AND. simp) then
      indentPrefx=repeat("| ",sectDepth(i))
    else if (present(spc) .AND. spc) then
      indentPrefx=repeat("  ",sectDepth(i))
    else
      indentPrefx=repeat("| ",sectDepth(i)-1)//"+"
    end if

  end function indentPrefx




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine startSubsect(units,genHdr,hdr,spcPrefx)
!
!   starts new subsection in standard output and/or log file
!
!   Arguments:
!     units                          output unit(s)
!     genHdr                         generic header string
!     hdr                            header strings for given output units
!     spcPrefx                       whether prefix should only consist of spaces
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     units, &
     i, &
     j
    logical :: &
     spcPrefx, &
     spP
    character(*) :: &
     genHdr, &
     hdr(:)
    optional :: &
     genHdr, &
     hdr, &
     spcPrefx
    intent(in) :: &
     hdr, &
     units, &
     genHdr, &
     spcPrefx

    spP=present(spcPrefx) .AND. spcPrefx
    j=0
    do i=0,UNIT%n-1
      if (ibits(units,i,1) == 1) then
        if (present(genHdr)) then
          call printMsg(2**i,genHdr,setFlg(FLG%spcPrfx,spP))
        else if (present(hdr)) then
          call inc(j)
          call printMsg(2**i,hdr(j),setFlg(FLG%spcPrfx,spP))
        end if
        call inc(sectDepth(i))
        if (i == I_UNIT%cons) then
          if (.NOT. usrQuiet) write(*,"(A)") indentPrefx(i,simp=.NOT.spP,spc=spP)
        else if (filOutput) then
          write(FORT_UNIT(i),"(A)") indentPrefx(i,simp=.NOT.spP,spc=spP)
        end if
      end if
    end do

  end subroutine startSubsect




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine endSubsect(units,genMsg,msg,spcPrefx)
!
!   ends current subsection in standard output and/or log file
!
!   Arguments:
!     units                          output unit(s)
!     genMsg                         generic message
!     msg                            messages for given output units
!     spcPrefx                       whether prefix should only consist of spaces
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     units, &
     i, &
     j
    logical :: &
     spcPrefx, &
     spP
    character(*) :: &
     genMsg, &
     msg(:)
    optional :: &
     genMsg, &
     msg, &
     spcPrefx
    intent(in) :: &
     msg, &
     units, &
     spcPrefx, &
     genMsg
    j=0

    spP=present(spcPrefx) .AND. spcPrefx
    do i=0,UNIT%n-1
      if (ibits(units,i,1) == 1 .AND. sectDepth(i) /= 0) then
        sectDepth(i)=sectDepth(i)-1
        if (present(genMsg)) then
          call printMsg(2**i,genMsg,setFlg([FLG%spcPrfx,FLG%simpPrfx],[spP,.NOT.spP]))
        else if (present(msg)) then
          call inc(j)
          call printMsg(2**i,msg(j),setFlg([FLG%spcPrfx,FLG%simpPrfx],[spP,.NOT.spP]))
        end if
        if (i == I_UNIT%cons) then
          if (.NOT. usrQuiet) write(*,"(A)") indentPrefx(i,simp=.NOT.spP,spc=spP)
        else if (filOutput) then
          write(FORT_UNIT(i),"(A)") indentPrefx(i,simp=.NOT.spP,spc=(present(spcPrefx) .AND. &
           spcPrefx))
        end if
      end if
    end do

  end subroutine endSubsect




  !==================================================[ CONSOLE AND LOG FILE ]=======================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fstStr(unit,desc)
!
!   Returns:                         first string after indentation prefix for standard output or log file,
!                                    e.g. containing description of a quantity
!
!   Arguments:
!     unit                           output unit
!     desc                           description
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     unit, &
     descLen, &
     pos1, &
     pos2, &
     pos3, &
     nSpcBefRuler
    character(*) :: &
     desc
    character(len_trim(adjustl(desc))) :: &
     adjDesc
    character(consTabStop-(1+sectDepth(unit)*CONS_TAB_WIDTH)) :: &
     fstStr
    intent(in) :: &
     unit, &
     desc

    adjDesc=trim(adjustl(desc))
    descLen=len(adjDesc)

    pos1=1+sectDepth(unit)*CONS_TAB_WIDTH
    pos2=pos1+descLen
    pos3=consTabStop

    if (pos1 > pos3-3) then
      pos1=pos3-3
      fstStr=".. "
    else if (pos2 > pos3-3) then
      pos2=pos3-1
      fstStr=adjDesc(1:pos2-pos1-2)//".. "
    else
      nSpcBefRuler=1+1+RULER_N_SPACES-mod(pos2+1,1+RULER_N_SPACES)
      fstStr=adjDesc//repeat(" ",nSpcBefRuler)//repeat("."//repeat(" ",RULER_N_SPACES), &
       floor(real(pos3-pos2-nSpcBefRuler-1)/(1+RULER_N_SPACES)))// &
       repeat(" ",mod(pos3-pos2-nSpcBefRuler-1,1+RULER_N_SPACES)+1)
    end if

  end function fstStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printInt(units,desc,intVal,advanc)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     intVal                         quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     intVal, &
     units
    logical :: &
     advanc
    character(*) :: &
     desc
    optional :: &
     advanc
    intent(in) :: &
     desc, &
     intVal, &
     advanc, &
     units

    call printQty(units,desc,trim(niceFmt(intVal)),(.NOT. present(advanc) .OR. advanc))

  end subroutine printInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printInt2(units,desc,intVal,advanc)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     intVal                         quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     units
    integer(i2b) :: &
     intVal
    logical :: &
     advanc
    character(*) :: &
     desc
    optional :: &
     advanc
    intent(in) :: &
     desc, &
     intVal, &
     advanc, &
     units

    call printQty(units,desc,trim(niceFmt(intVal)),(.NOT. present(advanc) .OR. advanc))

  end subroutine printInt2




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printInt8(units,desc,intVal,advanc)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     intVal                         quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     units
    integer(i8b) :: &
     intVal
    logical :: &
     advanc
    character(*) :: &
     desc
    optional :: &
     advanc
    intent(in) :: &
     desc, &
     intVal, &
     advanc, &
     units

    call printQty(units,desc,trim(niceFmt(intVal)),(.NOT. present(advanc) .OR. advanc))

  end subroutine printInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printReal(units,desc,realVal,advanc)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     realVal                        quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none

    integer :: &
     units
    real :: &
     realVal
    logical :: &
     advanc
    character(*) :: &
     desc
    optional :: &
     advanc
    intent(in) :: &
     desc, &
     realVal, &
     advanc, &
     units

    call printQty(units,desc,trim(niceFmt(realVal)),(.NOT. present(advanc) .OR. advanc))

  end subroutine printReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printDbl(units,desc,dblVal,advanc)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     dblVal                         quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     units
    real(dp) :: &
     dblVal
    logical :: &
     advanc
    character(*) :: &
     desc
    optional :: &
     advanc
    intent(in) :: &
     desc, &
     dblVal, &
     advanc, &
     units

    call printQty(units,desc,trim(niceFmt(dblVal)),(.NOT. present(advanc) .OR. advanc))

  end subroutine printDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printQP(units,desc,QPVal,advanc)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     QPVal                          quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     units
    real(qp) :: &
     QPVal
    logical :: &
     advanc
    character(*) :: &
     desc
    optional :: &
     advanc
    intent(in) :: &
     desc, &
     QPVal, &
     advanc, &
     units

    call printQty(units,desc,trim(niceFmt(QPVal)),(.NOT. present(advanc) .OR. advanc))

  end subroutine printQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printBool(units,desc,boolVal,advanc)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     boolVal                        quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     units
    logical :: &
     boolVal, &
     advanc
    character(*) :: &
     desc
    optional :: &
     advanc
    intent(in) :: &
     desc, &
     boolVal, &
     advanc, &
     units

    call printQty(units,desc,trim(merge(xtndStr(STR_TRUE,len(STR_FALSE)),xtndStr(STR_FALSE,len(STR_TRUE)),boolVal)), &
     (.NOT. present(advanc) .OR. advanc))

  end subroutine printBool




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printStr(units,desc,str,advanc)
!
!   prints a string, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     str                            string to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     descLen, &
     units, &
     i
    logical :: &
     advanc
    character(*) :: &
     desc, &
     str
    character(STR_LEN_LONGEST) :: &
     adjStr, &
     cplStr
    character(len_trim(adjustl(desc))) :: &
     adjDesc
    optional :: &
     advanc
    intent(in) :: &
     desc, &
     str, &
     advanc, &
     units


    adjDesc=trim(adjustl(desc))
    descLen=len(adjDesc)

    do i=0,UNIT%n-1

      if (isUndef(str)) then
        adjStr=STR_UNDEF
      else
        adjStr=adjustl(str)
      end if

      cplStr=indentPrefx(i)//rmTrailN(fstStr(i,desc),merge(1,0,adjStr(1:1) == "-"))//adjStr
      if (present(advanc) .AND. .NOT. advanc .AND. i == I_UNIT%cons .AND. .NOT. usrQuiet) then
        ! don't write to file in this case
        write(*,"(A)",advance="no") cplStr(1:consLineLen)//repeat(char(8),consLineLen)
        exit
      else if (ibits(units,i,1) == 1) then
        if (i == I_UNIT%cons) then
          if (.NOT. usrQuiet) write(*,"(A)") cplStr(1:consLineLen)
        else if (filOutput) then
          write(FORT_UNIT(i),"(A)") trim(cplStr)
        end if
      end if
    end do

  end subroutine printStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printIntvlInt(units,desc,intvl,withRge)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     intvl                          quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none

    integer :: &
     units
    logical :: &
     withRge
    character(*) :: &
     desc
    type(tyIntvlInt) :: &
     intvl
    optional :: &
     withRge
    intent(in) :: &
     units, &
     desc, &
     intvl, &
     withRge

    call startSubsect(units,desc)
    call printQty(units,uCaseFst(DESC_INTVL_L_BND),intvl%lBnd)
    call printQty(units,uCaseFst(DESC_INTVL_U_BND),intvl%uBnd)
    if (present(withRge) .AND. withRge) call printQty(units,uCaseFst(DESC_INTVL_RGE),intvl%width())
    call endSubsect(units)

  end subroutine printIntvlInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printIntvlReal(units,desc,intvl,withRge)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     intvl                          quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     units
    logical :: &
     withRge
    character(*) :: &
     desc
    type(tyIntvlReal) :: &
     intvl
    optional :: &
     withRge
    intent(in) :: &
     units, &
     desc, &
     intvl, &
     withRge

    call startSubsect(units,desc)
    call printQty(units,uCaseFst(DESC_INTVL_L_BND),intvl%lBnd)
    call printQty(units,uCaseFst(DESC_INTVL_U_BND),intvl%uBnd)
    if (present(withRge) .AND. withRge) call printQty(units,uCaseFst(DESC_INTVL_RGE),intvl%width())
    call endSubsect(units)

  end subroutine printIntvlReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printIntvlDbl(units,desc,intvl,withRge)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     intvl                          quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none

    integer :: &
     units
    logical :: &
     withRge
    character(*) :: &
     desc
    type(tyIntvlDbl) :: &
     intvl
    optional :: &
     withRge
    intent(in) :: &
     units, &
     desc, &
     intvl, &
     withRge

    call startSubsect(units,desc)
    call printQty(units,uCaseFst(DESC_INTVL_L_BND),intvl%lBnd)
    call printQty(units,uCaseFst(DESC_INTVL_U_BND),intvl%uBnd)
    if (present(withRge) .AND. withRge) call printQty(units,uCaseFst(DESC_INTVL_RGE),intvl%width())
    call endSubsect(units)

  end subroutine printIntvlDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printIntvlQP(units,desc,intvl,withRge)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     intvl                          quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none

    integer :: &
     units
    logical :: &
     withRge
    character(*) :: &
     desc
    type(tyIntvlQP) :: &
     intvl
    optional :: &
     withRge
    intent(in) :: &
     units, &
     desc, &
     intvl, &
     withRge

    call startSubsect(units,desc)
    call printQty(units,uCaseFst(DESC_INTVL_L_BND),intvl%lBnd)
    call printQty(units,uCaseFst(DESC_INTVL_U_BND),intvl%uBnd)
    if (present(withRge) .AND. withRge) call printQty(units,uCaseFst(DESC_INTVL_RGE),intvl%width())
    call endSubsect(units)

  end subroutine printIntvlQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printUncQtyDbl(units,desc,qty,stDev)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     qty                            quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none

    integer :: &
     units
    logical :: &
     stDev
    character(*) :: &
     desc
    type(tyUncQtyDbl) :: &
     qty
    optional :: &
     stDev
    intent(in) :: &
     units, &
     desc, &
     qty, &
     stDev

    call startSubsect(units,desc)
    call printQty(units,uCaseFst(DESC_UNC_QTY_VAL),qty%val)
    if (present(stDev) .AND. stDev) then
      call printQty(units,uCaseFst(DESC_STDEV),qty%unc)
    else
      call printQty(units,uCaseFst(DESC_UNC_QTY_UNC),qty%unc)
    end if
    call endSubsect(units)

  end subroutine printUncQtyDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printUncQtyQP(units,desc,qty,stDev)
!
!   prints a quantity, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     desc                           description of quantity
!     qty                            quantity to print
!     advanc                         whether newline should be appended to output
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none

    integer :: &
     units
    logical :: &
     stDev
    character(*) :: &
     desc
    type(tyUncQtyQP) :: &
     qty
    optional :: &
     stDev
    intent(in) :: &
     units, &
     desc, &
     qty, &
     stDev

    call startSubsect(units,desc)
    call printQty(units,uCaseFst(DESC_UNC_QTY_VAL),qty%val)
    if (present(stDev) .AND. stDev) then
      call printQty(units,uCaseFst(DESC_STDEV),qty%unc)
    else
      call printQty(units,uCaseFst(DESC_UNC_QTY_UNC),qty%unc)
    end if
    call endSubsect(units)

  end subroutine printUncQtyQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printMsg(units,msg,flags)
!
!   prints a message, obeying section level
!
!   Arguments:
!     units                          output unit(s)
!     msg                            message string
!     flags                          option flags
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     units, &
     i, &
     flags
    logical :: &
     advanc, &
     simpPrefx, &
     spcPrefx
    character(*) :: &
     msg
    character(STR_LEN_LONGEST) :: &
     cplStr
    optional :: &
     flags
    intent(in) :: &
     units, &
     msg, &
     flags

    advanc=.TRUE.
    simpPrefx=.FALSE.
    spcPrefx=.FALSE.
    if (present(flags)) then
      advanc=.NOT.getFlg(flags,FLG%noAdvc)
      simpPrefx=getFlg(flags,FLG%simpPrfx)
      spcPrefx=getFlg(flags,FLG%spcPrfx)
    end if

    do i=0,UNIT%n-1
      if (ibits(units,i,1) == 1) then
        cplStr=indentPrefx(i,simp=simpPrefx,spc=spcPrefx)//trim(msg)
        if (i == I_UNIT%cons) then
          if (.NOT. usrQuiet) then
            if (advanc) then
              write(*,"(A)") trim(cplStr)
            else
              write(*,"(A)",advance="no") cplStr(1:consLineLen)//repeat(char(8),consLineLen)
            end if
          end if
        else if (filOutput) then
          write(FORT_UNIT(i),"(A)") trim(cplStr)
        end if
      end if
    end do

  end subroutine printMsg




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine warn(msg)
!
!   prints a warning message, obeying section level
!
!   Arguments:
!     msg                            warning message
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*) :: &
     msg
    intent(in) :: &
     msg

    call printMsg(UNIT%cons+UNIT%log,"WARNING: "//trim(adjustl(msg)))

  end subroutine warn




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printHostProps
!
!   prints properties of host computer
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none

    call startSubsect(UNIT%cons+UNIT%log,"Host computer properties")
    call printQty(UNIT%cons+UNIT%log,"Name",hostName)
    call printQty(UNIT%cons+UNIT%log,"# CPUs",OMP_GET_NUM_PROCS())
    call endSubsect(UNIT%cons+UNIT%log)

  end subroutine printHostProps




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printPrgrs4(it,minIt,maxIt,nTot,reset)
!
!   prints information about the progress of some task
!
!   Arguments:
!     it                             current iteration index(es)
!     minIt                          minimum iteration index
!     maxIt                          maximum iteration index
!     nTot                           total # iterations
!     reset                          whether progress should be reset to zero
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     nTm1, &
     tm1Int
    integer(i4b) :: &
     it(:), &
     maxIt(:), &
     minIt(:), &
     nLps, &
     nT, &
     i, &
     nTot, &
     nCur=0
    real(dp) :: &
     curFrac, &
     curTm, &
     durRem, &
     tm1, &
     lstTm=huge(lstTm), &
     lstFrac=huge(lstFrac)
    logical :: &
     printETF=.FALSE., &
     reset
    optional :: &
     minIt, &
     maxIt, &
     nTot, &
     reset
    save :: &
     nCur, &
     nTm1, &
     tm1Int, &
     tm1, &
     lstTm, &
     lstFrac, &
     printETF
    intent(in) :: &
     it, &
     maxIt, &
     minIt, &
     nTot, &
     reset


    if (present(reset) .AND. reset) then

      nT=1
      nCur=0

    else if (present(nTot)) then

      nT=nTot
      call inc(nCur)

    else

      nLps=size(it,1)
      nT=product(maxIt-minIt+1)

      nCur=0
      do i=1,nLps-1
        nCur=nCur+(it(i)-minIt(i))*product(maxIt(i+1:nLps)-minIt(i+1:nLps)+1)
      end do
      nCur=nCur+it(nLps)

    end if

    curFrac=1._dp*nCur/nT
    curTm=omp_get_wtime()

    if (nCur == 1 .AND. nT == 1) then

      call printMsg(UNIT%cons+UNIT%log,"Done")
      nCur=0

    else

      if (curFrac < lstFrac) then    ! reset
        tm1=curTm
        tm1Int=time8()
        nTm1=nCur
        printETF=.FALSE.
      else
        if (curFrac >= 1._dp .AND. lstFrac == 1._dp) return
        durRem=(curTm-tm1)*(nT-nCur)/(nCur-nTm1)
        if ((curTm-lstTm > DISP_UPD_ETR_SELDOM_S .OR. durRem > DISP_UPD_ETR_LONG_S) .AND. .NOT. printETF) then
          call printQty(UNIT%cons,"Local date & time started",fmtDate(dateandtimeFromTime8(tm1Int),short=.TRUE.))
          printETF=.TRUE.
        end if
        if (printETF) then
          call printQty(UNIT%cons,"Estimated local date & time finished",fmtDate(dateandtimeFromTime8(tm1Int+ &
           nint((curTm-tm1)*nT/(nCur-nTm1))),short=.TRUE.),(curFrac == 1._dp))
        else
          call printQty(UNIT%cons,"Estimated time remaining [h:m:s]",fmtTime(nint(durRem),hrs3=.TRUE.),(curFrac == 1._dp))
        end if
        if (curFrac == 1._dp) then
          call printMsg(UNIT%log,"Done")
          nCur=0
        end if
      end if

    end if

    lstFrac=curFrac
    lstTm=curTm
    if (.NOT. present(nTot)) nCur=0

  end subroutine printPrgrs4




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine printPrgrs8(it,minIt,maxIt,nTot,reset)
!
!   prints information about the progress of some task
!
!   Arguments:
!     it                             current iteration index(es)
!     minIt                          minimum iteration index
!     maxIt                          maximum iteration index
!     nTot                           total # iterations
!     reset                          whether progress should be reset to zero
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     tm1Int
    integer(i8b) :: &
     it(:), &
     maxIt(:), &
     minIt(:), &
     nLps, &
     nT, &
     i, &
     nTot, &
     nTm1, &
     nCur=0
    real(dp) :: &
     curFrac, &
     curTm, &
     durRem, &
     tm1, &
     lstTm=huge(lstTm), &
     lstFrac=huge(lstFrac)
    logical :: &
     printETF=.FALSE., &
     reset
    optional :: &
     minIt, &
     maxIt, &
     nTot, &
     reset
    save :: &
     nCur, &
     nTm1, &
     tm1Int, &
     tm1, &
     lstTm, &
     lstFrac, &
     printETF
    intent(in) :: &
     it, &
     maxIt, &
     minIt, &
     nTot, &
     reset


    if (present(reset) .AND. reset) then

      nT=1
      nCur=0

    else if (present(nTot)) then

      nT=nTot
      call inc(nCur)

    else

      nLps=size(it,1)
      nT=product(maxIt-minIt+1)

      nCur=0
      do i=1,nLps-1
        nCur=nCur+(it(i)-minIt(i))*product(maxIt(i+1:nLps)-minIt(i+1:nLps)+1)
      end do
      nCur=nCur+it(nLps)

    end if

    curFrac=1._dp*nCur/nT
    curTm=omp_get_wtime()

    if (nCur == 1 .AND. nT == 1) then

      call printMsg(UNIT%cons+UNIT%log,"Done")
      nCur=0

    else

      if (curFrac < lstFrac) then    ! reset
        tm1=curTm
        tm1Int=time8()
        nTm1=nCur
        printETF=.FALSE.
      else
        if (curFrac >= 1._dp .AND. lstFrac == 1._dp) return
        durRem=(curTm-tm1)*(nT-nCur)/(nCur-nTm1)
        if ((curTm-lstTm > DISP_UPD_ETR_SELDOM_S .OR. durRem > DISP_UPD_ETR_LONG_S) .AND. .NOT. printETF) then
          call printQty(UNIT%cons,"Local date & time started",fmtDate(dateandtimeFromTime8(tm1Int),short=.TRUE.))
          printETF=.TRUE.
        end if
        if (printETF) then
          call printQty(UNIT%cons,"Estimated local date & time finished",fmtDate(dateandtimeFromTime8(tm1Int+ &
           nint((curTm-tm1)*nT/(nCur-nTm1))),short=.TRUE.),(curFrac == 1._dp))
        else
          call printQty(UNIT%cons,"Estimated time remaining [h:m:s]",fmtTime(nint(durRem),hrs3=.TRUE.),(curFrac == 1._dp))
        end if
        if (curFrac == 1._dp) then
          call printMsg(UNIT%log,"Done")
          nCur=0
        end if
      end if

    end if

    lstFrac=curFrac
    lstTm=curTm
    if (.NOT. present(nTot)) nCur=0

  end subroutine printPrgrs8




  !=======================================================[ OTHER FILES ]===========================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function filNmPrefx(sysNm,dateStartPrg)
!
!   Returns:                         generic file-name prefix
!
!   Arguments:
!     sysNm                          identifier of observed system
!     dateStartPrg                   date and time of start of BASE run
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     dateStartPrg(8)
    character(120) :: &
     filNmPrefx
    character(*) :: &
     sysNm
    intent(in) :: &
     sysNm, &
     dateStartPrg

    filNmPrefx=trim(trim(fmtFilNm(sysNm)))//"_"//fmtDate(dateStartPrg,filNm=.TRUE.)

  end function filNmPrefx




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtFilNm(s)
!
!   Returns:                         string formatted for use in a file name
!
!   Arguments:
!     s                              original string
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i, &
     l
    character(*) :: &
     s
    character(max(len(STR_UNDEF_FIL_NM),len_trim(adjustl(s)))) :: &
     fmtFilNm
    intent(in) :: &
     s

    if (isUndef(s)) then
      fmtFilNm=STR_UNDEF_FIL_NM
    else
      fmtFilNm=adjustl(s)
    end if

    l=len_trim(fmtFilNm)

    do i=1,l
      if (scan(FIL_NM_ILLEGAL_CHARS,fmtFilNm(i:i)) > 0) fmtFilNm(i:i)="_"
    end do

  end function fmtFilNm




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function readCSVArrDbl(path,arr,nL)
!
!   Returns:                         whether a given CSV file has no data left, after reading values from it
!
!   Arguments:
!     path                           file path
!     arr                            array to fill from file
!     nL                             # lines according to arr
!
!   arr IS CONTIGUOUS (as of rev. 115)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i, &
     j, &
     readStat, &
     nL, &
     nC
    real(dp) :: &
     arr(:,:)
    logical :: &
     readCSVArrDbl
    character(*) :: &
     path
    character(STR_LEN_LONG) :: &
     lin
    intent(in) :: &
     path
    intent(inout) :: &
     arr
    intent(out) :: &
     nL

    call tryOpenFilRW("",trim(path),UN_TMP,"old","read")

    nL=size(arr,1)
    nC=size(arr,2)

    i=0
    do
      read (UN_TMP,"(A)",iostat=readStat) lin
      if (readStat /= 0) then
        readCSVArrDbl=.TRUE.         ! ok
        exit
      else if (beginsWith(lin,COMMENT_PREFX)) then
        cycle
      end if
      lin=adjustl(lin)
      call inc(i)
      do j=1,nC
        read (lin,*) arr(i,:)
      end do
      if (i == nL) then
        readCSVArrDbl=(readStat /= 0) ! .TRUE. if reading in file has just been finished, .FALSE. if there are more values
        exit
      end if

    end do

    nL=i

  end function readCSVArrDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeCSVIntReal(path,arrX,arrY,arrZ,hdrLins,hdrArr,leftArr,vSpc)
!
!   writes values to CSV file
!
!   Arguments:
!     path                           file path
!     arrX                           values for first column
!     arrY                           values for second column
!     arrZ                           values for third column
!     hdrLins                        header lines
!     hdrArr                         header fields
!     leftArr                        fields to be put left of first column with values
!     vSpc                           whether blank lines should be inserted at certain positions (used for 2D KDE points)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     arrX(:), &
     s, &
     i, &
     sp, &
     n, &
     j, &
     l
    real :: &
     arrY(:), &
     arrZ(:)
    logical :: &
     vSpc
    character(*) :: &
     path, &
     hdrLins(:), &
     hdrArr(:,:), &
     leftArr(:,:)
    character(STR_LEN_LONGEST) :: &
     curL
    optional :: &
     hdrLins, &
     hdrArr, &
     leftArr, &
     arrZ, &
     vSpc
    intent(in) :: &
     path, &
     arrX, &
     arrY, &
     arrZ, &
     hdrLins, &
     hdrArr, &
     leftArr, &
     vSpc

    s=min(size(arrX),size(arrY))
    if (present(arrZ)) s=min(s,size(arrZ))
    if (present(vSpc) .AND. vSpc) then
      sp=int(sqrt(1.*s))
    else
      sp=0
    end if

    call tryOpenFilRW("CSV",path,UN_PLT_CSV,"replace","write")

    if (present(hdrLins)) then
      do i=1,ubound(hdrLins,1)
        write(UN_PLT_CSV,"(A)") trim(hdrLins(i))
      end do
    end if

    if (present(hdrArr)) then
      do i=1,ubound(hdrArr,1)
        curL=""
        do j=1,ubound(hdrArr,2)
          l=merge(len(CSV_SEP),0,j > 1)
          curL=trim(curL)//CSV_SEP(1:l)//trim(hdrArr(i,j))
        end do
        write(UN_PLT_CSV,"(A)") trim(curL)
      end do
    end if

    n=0
    do i=1,s
      curL=""
      if (present(leftArr)) then
        do j=1,ubound(leftArr,2)
          l=merge(len(CSV_SEP),0,j > 1)
          curL=trim(curL)//CSV_SEP(1:l)//trim(leftArr(i,j))
        end do
      end if
      l=merge(len(CSV_SEP),0,present(leftArr))
      if (present(arrZ)) then
        curL=trim(curL)//CSV_SEP(1:l)//trim(niceFmt(arrX(i)))//CSV_SEP//trim(niceFmt(arrY(i)))//CSV_SEP//trim(niceFmt(arrZ(i)))
      else
        curL=trim(curL)//CSV_SEP(1:l)//trim(niceFmt(arrX(i)))//CSV_SEP//trim(niceFmt(arrY(i)))
      end if
      write(UN_PLT_CSV,"(A)") trim(curL)
      call inc(n)
      if (n == sp) then
        write(UN_PLT_CSV,"(A)")
        n=0
      end if
    end do
    close(UN_PLT_CSV)

  end subroutine writeCSVIntReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeCSVIntDbl(path,arrX,arrY,arrZ,hdrLins,hdrArr,leftArr,vSpc)
!
!   writes values to CSV file
!
!   Arguments:
!     path                           file path
!     arrX                           values for first column
!     arrY                           values for second column
!     arrZ                           values for third column
!     hdrLins                        header lines
!     hdrArr                         header fields
!     leftArr                        fields to be put left of first column with values
!     vSpc                           whether blank lines should be inserted at certain positions (used for 2D KDE points)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     arrX(:), &
     s, &
     i, &
     sp, &
     n, &
     j, &
     l
    real(dp) :: &
     arrY(:), &
     arrZ(:)
    logical :: &
     vSpc
    character(*) :: &
     path, &
     hdrLins(:), &
     hdrArr(:,:), &
     leftArr(:,:)
    character(STR_LEN_LONGEST) :: &
     curL
    optional :: &
     hdrLins, &
     hdrArr, &
     leftArr, &
     arrZ, &
     vSpc
    intent(in) :: &
     path, &
     arrX, &
     arrY, &
     arrZ, &
     hdrLins, &
     hdrArr, &
     leftArr, &
     vSpc

    s=min(size(arrX),size(arrY))
    if (present(arrZ)) s=min(s,size(arrZ))
    if (present(vSpc) .AND. vSpc) then
      sp=int(sqrt(1.*s))
    else
      sp=0
    end if

    call tryOpenFilRW("CSV",path,UN_PLT_CSV,"replace","write")

    if (present(hdrLins)) then
      do i=1,ubound(hdrLins,1)
        write(UN_PLT_CSV,"(A)") trim(hdrLins(i))
      end do
    end if

    if (present(hdrArr)) then
      do i=1,ubound(hdrArr,1)
        curL=""
        do j=1,ubound(hdrArr,2)
          l=merge(len(CSV_SEP),0,j > 1)
          curL=trim(curL)//CSV_SEP(1:l)//trim(hdrArr(i,j))
        end do
        write(UN_PLT_CSV,"(A)") trim(curL)
      end do
    end if

    n=0
    do i=1,s
      curL=""
      if (present(leftArr)) then
        do j=1,ubound(leftArr,2)
          l=merge(len(CSV_SEP),0,j > 1)
          curL=trim(curL)//CSV_SEP(1:l)//trim(leftArr(i,j))
        end do
      end if
      l=merge(len(CSV_SEP),0,present(leftArr))
      if (present(arrZ)) then
        curL=trim(curL)//CSV_SEP(1:l)//trim(niceFmt(arrX(i)))//CSV_SEP//trim(niceFmt(arrY(i)))//CSV_SEP//trim(niceFmt(arrZ(i)))
      else
        curL=trim(curL)//CSV_SEP(1:l)//trim(niceFmt(arrX(i)))//CSV_SEP//trim(niceFmt(arrY(i)))
      end if
      write(UN_PLT_CSV,"(A)") trim(curL)
      call inc(n)
      if (n == sp) then
        write(UN_PLT_CSV,"(A)")
        n=0
      end if
    end do
    close(UN_PLT_CSV)

  end subroutine writeCSVIntDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeCSVInt8Dbl(path,arrX,arrY,arrZ,hdrLins,hdrArr,leftArr,vSpc)
!
!   writes values to CSV file
!
!   Arguments:
!     path                           file path
!     arrX                           values for first column
!     arrY                           values for second column
!     arrZ                           values for third column
!     hdrLins                        header lines
!     hdrArr                         header fields
!     leftArr                        fields to be put left of first column with values
!     vSpc                           whether blank lines should be inserted at certain positions (used for 2D KDE points)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     j, &
     l
    integer(i8b) :: &
     sp, &
     n, &
     s, &
     i, &
     arrX(:)
    real(dp) :: &
     arrY(:), &
     arrZ(:)
    logical :: &
     vSpc
    character(*) :: &
     path, &
     hdrLins(:), &
     hdrArr(:,:), &
     leftArr(:,:)
    character(STR_LEN_LONGEST) :: &
     curL
    optional :: &
     hdrLins, &
     hdrArr, &
     leftArr, &
     arrZ, &
     vSpc
    intent(in) :: &
     path, &
     arrX, &
     arrY, &
     arrZ, &
     hdrLins, &
     hdrArr, &
     leftArr, &
     vSpc

    s=min(size(arrX),size(arrY))
    if (present(arrZ)) s=min(s,int(size(arrZ),i8b))
    if (present(vSpc) .AND. vSpc) then
      sp=int(sqrt(1.*s))
    else
      sp=0
    end if

    call tryOpenFilRW("CSV",path,UN_PLT_CSV,"replace","write")

    if (present(hdrLins)) then
      do i=1,ubound(hdrLins,1)
        write(UN_PLT_CSV,"(A)") trim(hdrLins(i))
      end do
    end if

    if (present(hdrArr)) then
      do i=1,ubound(hdrArr,1)
        curL=""
        do j=1,ubound(hdrArr,2)
          l=merge(len(CSV_SEP),0,j > 1)
          curL=trim(curL)//CSV_SEP(1:l)//trim(hdrArr(i,j))
        end do
        write(UN_PLT_CSV,"(A)") trim(curL)
      end do
    end if

    n=0
    do i=1,s
      curL=""
      if (present(leftArr)) then
        do j=1,ubound(leftArr,2)
          l=merge(len(CSV_SEP),0,j > 1)
          curL=trim(curL)//CSV_SEP(1:l)//trim(leftArr(i,j))
        end do
      end if
      l=merge(len(CSV_SEP),0,present(leftArr))
      if (present(arrZ)) then
        curL=trim(curL)//CSV_SEP(1:l)//trim(niceFmt(arrX(i)))//CSV_SEP//trim(niceFmt(arrY(i)))//CSV_SEP//trim(niceFmt(arrZ(i)))
      else
        curL=trim(curL)//CSV_SEP(1:l)//trim(niceFmt(arrX(i)))//CSV_SEP//trim(niceFmt(arrY(i)))
      end if
      write(UN_PLT_CSV,"(A)") trim(curL)
      call inc(n)
      if (n == sp) then
        write(UN_PLT_CSV,"(A)")
        n=0
      end if
    end do
    close(UN_PLT_CSV)

  end subroutine writeCSVInt8Dbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeCSVInt8QP(path,arrX,arrY,arrZ,hdrLins,hdrArr,leftArr,vSpc)
!
!   writes values to CSV file
!
!   Arguments:
!     path                           file path
!     arrX                           values for first column
!     arrY                           values for second column
!     arrZ                           values for third column
!     hdrLins                        header lines
!     hdrArr                         header fields
!     leftArr                        fields to be put left of first column with values
!     vSpc                           whether blank lines should be inserted at certain positions (used for 2D KDE points)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     j, &
     l
    integer(i8b) :: &
     sp, &
     n, &
     s, &
     i, &
     arrX(:)
    real(qp) :: &
     arrY(:), &
     arrZ(:)
    logical :: &
     vSpc
    character(*) :: &
     path, &
     hdrLins(:), &
     hdrArr(:,:), &
     leftArr(:,:)
    character(STR_LEN_LONGEST) :: &
     curL
    optional :: &
     hdrLins, &
     hdrArr, &
     leftArr, &
     arrZ, &
     vSpc
    intent(in) :: &
     path, &
     arrX, &
     arrY, &
     arrZ, &
     hdrLins, &
     hdrArr, &
     leftArr, &
     vSpc

    s=min(size(arrX),size(arrY))
    if (present(arrZ)) s=min(s,int(size(arrZ),i8b))
    if (present(vSpc) .AND. vSpc) then
      sp=int(sqrt(1.*s))
    else
      sp=0
    end if

    call tryOpenFilRW("CSV",path,UN_PLT_CSV,"replace","write")

    if (present(hdrLins)) then
      do i=1,ubound(hdrLins,1)
        write(UN_PLT_CSV,"(A)") trim(hdrLins(i))
      end do
    end if

    if (present(hdrArr)) then
      do i=1,ubound(hdrArr,1)
        curL=""
        do j=1,ubound(hdrArr,2)
          l=merge(len(CSV_SEP),0,j > 1)
          curL=trim(curL)//CSV_SEP(1:l)//trim(hdrArr(i,j))
        end do
        write(UN_PLT_CSV,"(A)") trim(curL)
      end do
    end if

    n=0
    do i=1,s
      curL=""
      if (present(leftArr)) then
        do j=1,ubound(leftArr,2)
          l=merge(len(CSV_SEP),0,j > 1)
          curL=trim(curL)//CSV_SEP(1:l)//trim(leftArr(i,j))
        end do
      end if
      l=merge(len(CSV_SEP),0,present(leftArr))
      if (present(arrZ)) then
        curL=trim(curL)//CSV_SEP(1:l)//trim(niceFmt(arrX(i)))//CSV_SEP//trim(niceFmt(arrY(i)))//CSV_SEP//trim(niceFmt(arrZ(i)))
      else
        curL=trim(curL)//CSV_SEP(1:l)//trim(niceFmt(arrX(i)))//CSV_SEP//trim(niceFmt(arrY(i)))
      end if
      write(UN_PLT_CSV,"(A)") trim(curL)
      call inc(n)
      if (n == sp) then
        write(UN_PLT_CSV,"(A)")
        n=0
      end if
    end do
    close(UN_PLT_CSV)

  end subroutine writeCSVInt8QP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeCSVRealReal(path,arrX,arrY,arrZ,hdrLins,hdrArr,leftArr,vSpc)
!
!   writes values to CSV file
!
!   Arguments:
!     path                           file path
!     arrX                           values for first column
!     arrY                           values for second column
!     arrZ                           values for third column
!     hdrLins                        header lines
!     hdrArr                         header fields
!     leftArr                        fields to be put left of first column with values
!     vSpc                           whether blank lines should be inserted at certain positions (used for 2D KDE points)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     s, &
     i, &
     sp, &
     n, &
     j, &
     l
    real :: &
     arrX(:), &
     arrY(:), &
     arrZ(:)
    logical :: &
     vSpc
    character(*) :: &
     path, &
     hdrLins(:), &
     hdrArr(:,:), &
     leftArr(:,:)
    character(STR_LEN_LONGEST) :: &
     curL
    optional :: &
     hdrLins, &
     hdrArr, &
     leftArr, &
     arrZ, &
     vSpc
    intent(in) :: &
     path, &
     arrX, &
     arrY, &
     arrZ, &
     hdrLins, &
     hdrArr, &
     leftArr, &
     vSpc

    s=min(size(arrX),size(arrY))
    if (present(arrZ)) s=min(s,size(arrZ))
    if (present(vSpc) .AND. vSpc) then
      sp=int(sqrt(1.*s))
    else
      sp=0
    end if

    call tryOpenFilRW("CSV",path,UN_PLT_CSV,"replace","write")

    if (present(hdrLins)) then
      do i=1,ubound(hdrLins,1)
        write(UN_PLT_CSV,"(A)") trim(hdrLins(i))
      end do
    end if

    if (present(hdrArr)) then
      do i=1,ubound(hdrArr,1)
        curL=""
        do j=1,ubound(hdrArr,2)
          l=merge(len(CSV_SEP),0,j > 1)
          curL=trim(curL)//CSV_SEP(1:l)//trim(hdrArr(i,j))
        end do
        write(UN_PLT_CSV,"(A)") trim(curL)
      end do
    end if

    n=0
    do i=1,s
      curL=""
      if (present(leftArr)) then
        do j=1,ubound(leftArr,2)
          l=merge(len(CSV_SEP),0,j > 1)
          curL=trim(curL)//CSV_SEP(1:l)//trim(leftArr(i,j))
        end do
      end if
      l=merge(len(CSV_SEP),0,present(leftArr))
      if (present(arrZ)) then
        curL=trim(curL)//CSV_SEP(1:l)//trim(niceFmt(arrX(i)))//CSV_SEP//trim(niceFmt(arrY(i)))//CSV_SEP//trim(niceFmt(arrZ(i)))
      else
        curL=trim(curL)//CSV_SEP(1:l)//trim(niceFmt(arrX(i)))//CSV_SEP//trim(niceFmt(arrY(i)))
      end if
      write(UN_PLT_CSV,"(A)") trim(curL)
      call inc(n)
      if (n == sp) then
        write(UN_PLT_CSV,"(A)")
        n=0
      end if
    end do
    close(UN_PLT_CSV)

  end subroutine writeCSVRealReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeCSVDblDbl(path,arrX,arrY,arrZ,hdrLins,hdrArr,leftArr,vSpc)
!
!   writes values to CSV file
!
!   Arguments:
!     path                           file path
!     arrX                           values for first column
!     arrY                           values for second column
!     arrZ                           values for third column
!     hdrLins                        header lines
!     hdrArr                         header fields
!     leftArr                        fields to be put left of first column with values
!     vSpc                           whether blank lines should be inserted at certain positions (used for 2D KDE points)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     s, &
     i, &
     sp, &
     n, &
     j, &
     l
    real(dp) :: &
     arrX(:), &
     arrY(:), &
     arrZ(:)
    logical :: &
     vSpc
    character(*) :: &
     path, &
     hdrLins(:), &
     hdrArr(:,:), &
     leftArr(:,:)
    character(STR_LEN_LONGEST) :: &
     curL
    optional :: &
     hdrLins, &
     hdrArr, &
     leftArr, &
     arrZ, &
     vSpc
    intent(in) :: &
     path, &
     arrX, &
     arrY, &
     arrZ, &
     hdrLins, &
     hdrArr, &
     leftArr, &
     vSpc

    s=min(size(arrX),size(arrY))
    if (present(arrZ)) s=min(s,size(arrZ))
    if (present(vSpc) .AND. vSpc) then
      sp=int(sqrt(1.*s))
    else
      sp=0
    end if

    call tryOpenFilRW("CSV",path,UN_PLT_CSV,"replace","write")

    if (present(hdrLins)) then
      do i=1,ubound(hdrLins,1)
        write(UN_PLT_CSV,"(A)") trim(hdrLins(i))
      end do
    end if

    if (present(hdrArr)) then
      do i=1,ubound(hdrArr,1)
        curL=""
        do j=1,ubound(hdrArr,2)
          l=merge(len(CSV_SEP),0,j > 1)
          curL=trim(curL)//CSV_SEP(1:l)//trim(hdrArr(i,j))
        end do
        write(UN_PLT_CSV,"(A)") trim(curL)
      end do
    end if

    n=0
    do i=1,s
      curL=""
      if (present(leftArr)) then
        do j=1,ubound(leftArr,2)
          l=merge(len(CSV_SEP),0,j > 1)
          curL=trim(curL)//CSV_SEP(1:l)//trim(leftArr(i,j))
        end do
      end if
      l=merge(len(CSV_SEP),0,present(leftArr))
      if (present(arrZ)) then
        curL=trim(curL)//CSV_SEP(1:l)//trim(niceFmt(arrX(i)))//CSV_SEP//trim(niceFmt(arrY(i)))//CSV_SEP//trim(niceFmt(arrZ(i)))
      else
        curL=trim(curL)//CSV_SEP(1:l)//trim(niceFmt(arrX(i)))//CSV_SEP//trim(niceFmt(arrY(i)))
      end if
      write(UN_PLT_CSV,"(A)") trim(curL)
      call inc(n)
      if (n == sp) then
        write(UN_PLT_CSV,"(A)")
        n=0
      end if
    end do
    close(UN_PLT_CSV)

  end subroutine writeCSVDblDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeArrCSVDbl(path,arr,hdrLins,hdrArr,leftArr)
!
!   writes values to CSV file
!
!   Arguments:
!     path                           file path
!     arr                            values for all columns
!     hdrLins                        header lines
!     hdrArr                         header fields
!     leftArr                        fields to be put left of first column with values
!
!   arr IS CONTIGUOUS (as of rev. 115)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     sX, &
     sY, &
     iX, &
     iY, &
     curLen, &
     curPos, &
     i, &
     sHX, &
     sHY, &
     sLX
    real(dp) :: &
     arr(:,:)
    character(*) :: &
     path, &
     hdrLins(:), &
     hdrArr(:,:), &
     leftArr(:,:)
    character(STR_LEN_LONGEST) :: &
     curL
    character(STR_LEN_STD) :: &
     cTmp, &
     cArr(size(arr,1),size(arr,2))
    optional :: &
     hdrLins, &
     hdrArr, &
     leftArr
    intent(in) :: &
     path, &
     arr, &
     hdrLins, &
     hdrArr, &
     leftArr

    sX=size(arr,2)
    sY=size(arr,1)

    do iY=1,sY
      do iX=1,sX
        cArr(iY,iX)=niceFmt(arr(iY,iX))
      end do
    end do
    
    if (present(hdrLins)) then
      if (present(hdrArr)) then
        if (present(leftArr)) then
          call writeArrCSV(path,cArr,hdrLins,hdrArr,leftArr)
        else
          call writeArrCSV(path,cArr,hdrLins,hdrArr)
        end if
      else
        if (present(leftArr)) then
          call writeArrCSV(path,cArr,hdrLins,leftArr=leftArr)
        else
          call writeArrCSV(path,cArr,hdrLins)
        end if
      end if
    else
      if (present(hdrArr)) then
        if (present(leftArr)) then
          call writeArrCSV(path,cArr,hdrArr=hdrArr,leftArr=leftArr)
        else
          call writeArrCSV(path,cArr,hdrArr=hdrArr)
        end if
      else
        if (present(leftArr)) then
          call writeArrCSV(path,cArr,leftArr=leftArr)
        else
          call writeArrCSV(path,cArr)
        end if
      end if
    end if

  end subroutine writeArrCSVDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine writeArrCSVStr(path,arr,hdrLins,hdrArr,leftArr)
!
!   writes values to CSV file
!
!   Arguments:
!     path                           file path
!     arr                            values for all columns
!     hdrLins                        header lines
!     hdrArr                         header fields
!     leftArr                        fields to be put left of first column with values
!
!   arr IS CONTIGUOUS (as of rev. 115)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     sX, &
     sY, &
     iX, &
     iY, &
     curLen, &
     curPos, &
     i, &
     sHX, &
     sHY, &
     sLX
    character(*) :: &
     path, &
     arr(:,:), &
     hdrLins(:), &
     hdrArr(:,:), &
     leftArr(:,:)
    character(STR_LEN_LONGEST) :: &
     curL
    character(STR_LEN_STD) :: &
     cTmp
    optional :: &
     hdrLins, &
     hdrArr, &
     leftArr
    intent(in) :: &
     path, &
     arr, &
     hdrLins, &
     hdrArr, &
     leftArr

    sX=size(arr,2)
    sY=size(arr,1)

    call tryOpenFilRW("CSV",path,UN_PLT_CSV,"replace","write")

    if (present(hdrLins)) then
      do i=1,ubound(hdrLins,1)
        write(UN_PLT_CSV,"(A)") trim(hdrLins(i))
      end do
    end if

    if (present(hdrArr)) then
      sHX=size(hdrArr,2)
      sHY=size(hdrArr,1)
      do iY=1,sHY
        curL=""
        curPos=1
        do iX=1,sHX
          if (iX == 1) then
            cTmp=trim(hdrArr(iY,iX))
          else
            cTmp=CSV_SEP//trim(hdrArr(iY,iX))
          end if
          curLen=len_trim(cTmp)
          curL(curPos:curPos+curLen-1)=trim(cTmp)
          curPos=curPos+curLen
        end do
        write(UN_PLT_CSV,"(A)") trim(curL)
      end do
    end if

    if (present(leftArr)) sLX=size(leftArr,2)

    do iY=1,sY
      curL=""
      curPos=1
      if (present(leftArr)) then
        do iX=1,sLX
          if (iX == 1) then
            cTmp=trim(leftArr(iY,iX))
          else
            cTmp=CSV_SEP//trim(leftArr(iY,iX))
          end if
          curLen=len_trim(cTmp)
          curL(curPos:curPos+curLen-1)=trim(cTmp)
          curPos=curPos+curLen
        end do
      end if
      do iX=1,sX
        if (isDef(arr(iY,iX))) then
          cTmp=arr(iY,iX)
        else
          cTmp=STR_UNDEF
        end if
        if (iX > 1 .OR. present(leftArr)) then
          cTmp=CSV_SEP//cTmp
        end if
        curLen=len_trim(cTmp)
        curL(curPos:curPos+curLen-1)=trim(cTmp)
        curPos=curPos+curLen
      end do
      write(UN_PLT_CSV,"(A)") trim(curL)
    end do
    close(UN_PLT_CSV)

  end subroutine writeArrCSVStr




  !======================================================[ INTERACTIVITY ]==========================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readIntQty(desc,val)
!
!   reads a quantity from standard input after printing its description to standard output
!
!   Arguments:
!     desc                           description of quantity
!     val                            value read
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     val
    character(*) :: &
     desc
    intent(in) :: &
     desc
    intent(out) :: &
     val

    write(*,"(A)",advance="no") indentPrefx(I_UNIT%cons)//fstStr(I_UNIT%cons,desc)
    read *,val

  end subroutine readIntQty




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readRealQty(desc,val)
!
!   reads a quantity from standard input after printing its description to standard output
!
!   Arguments:
!     desc                           description of quantity
!     val                            value read
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     val
    character(*) :: &
     desc
    intent(in) :: &
     desc
    intent(out) :: &
     val

    write(*,"(A)",advance="no") indentPrefx(I_UNIT%cons)//fstStr(I_UNIT%cons,desc)
    read *,val

  end subroutine readRealQty




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readDblQty(desc,val)
!
!   reads a quantity from standard input after printing its description to standard output
!
!   Arguments:
!     desc                           description of quantity
!     val                            value read
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     val
    character(*) :: &
     desc
    intent(in) :: &
     desc
    intent(out) :: &
     val

    write(*,"(A)",advance="no") indentPrefx(I_UNIT%cons)//fstStr(I_UNIT%cons,desc)
    read *,val

  end subroutine readDblQty




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readBoolQty(desc,val)
!
!   reads a quantity from standard input after printing its description to standard output
!
!   Arguments:
!     desc                           description of quantity
!     val                            value read
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     val
    character :: &
     ans
    character(*) :: &
     desc
    intent(in) :: &
     desc
    intent(out) :: &
     val

    write(*,"(A)",advance="no") indentPrefx(I_UNIT%cons)//fstStr(I_UNIT%cons,desc)
    read *,ans
    val=scan(CHARS_TRUE,ans) > 0

  end subroutine readBoolQty




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine readStrQty(desc,val)
!
!   reads a quantity from standard input after printing its description to standard output
!
!   Arguments:
!     desc                           description of quantity
!     val                            value read
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*) :: &
     desc, &
     val
    intent(in) :: &
     desc
    intent(out) :: &
     val

    write(*,"(A)",advance="no") indentPrefx(I_UNIT%cons)//fstStr(I_UNIT%cons,desc)
    read *,val

  end subroutine readStrQty




end module
