
                                                          program main

!                                                       The main program




  use baseRun
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !========================================================[ VARIABLES ]============================================================

  integer :: &
   i

  real(dp) :: &
   modeTmp(N_L_PARS_MAX)

  character(STR_LEN_STD) :: &
   hdr


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ EXECUTABLE PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––



!   · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
!
!     INITIALIZATIONS
!
!   · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·


  dbgSwitch=0
  call ini



  ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
  !
  !   PREPARATION OF ANALYSIS
  !
  ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·


  !  PARSE OPTIONS FROM CONFIG FILE AND/OR COMMAND LINE

  call parseOpts


  ! PRINT HELP SCREEN

  if (help) then
    call printHelp
    call exit
  else if (helpParsDQs) then
    call printParsDQsTbl
    call exit
  end if


  ! PARSE INPUT DATA

  if (.NOT. usrGenerateData .AND. dtProps%nSets > 0) call parseData ! error is cast later if original data not supplied, but info about posterior samples is printed before; thus we don't cast the error here


  ! OPEN LOG FILE

  runIdent=filNmPrefx(dtProps%sysNm,dateStartPrg)
  if (filOutput) then
    filNmLog=trim(outDir)//trim(runIdent)//".log"
    call tryOpenFilRW("log",filNmLog,FORT_UNIT(I_UNIT%log),"replace","write") ! log file
  end if


  ! SOME FIRST MESSAGES

  call printMsg(UNIT%cons+UNIT%log,"BASE - "//BASE_LONG//trim(niceFmt(BASE_REV_NR)))
  call startSubsect(UNIT%cons+UNIT%log,uCaseFst(AUTHOR_STR))

  if (nConfFilLins > 0) then
    call startSubsect(UNIT%cons+UNIT%log,"Config file")
    do i=1,nConfFilLins
      call printMsg(UNIT%cons+UNIT%log,trim(adjustl(confFilLins(i))))
    end do
    call endSubsect(UNIT%cons+UNIT%log)
  end if

  call startSubsect(UNIT%cons+UNIT%log,"Command line")
  call printMsg(UNIT%cons+UNIT%log,trim(adjustl(cmdLin)))
  call endSubsect(UNIT%cons+UNIT%log)


  ! USER COMMENT

  if (isDef(usrComment)) then
    call startSubsect(UNIT%cons+UNIT%log,"User comment")
    call printMsg(UNIT%cons+UNIT%log,trim(adjustl(usrComment)))
    call endSubsect(UNIT%cons+UNIT%log)
  end if


  ! LOCAL DATE & TIME, HOST INFO

  call printQty(UNIT%cons+UNIT%log,"Local date & time",fmtDate(dateStartPrg))

  call hostNm(hostName)
  call printHostProps

  
  ! GROUP AND EVALUATE DATA
  
  if (.NOT. usrGenerateData .AND. dtProps%nSets > 0) then
    call printMsg(UNIT%cons+UNIT%log,"All data within each set have been sorted by time")
    call groupData
    call evalDtProps
  end if


  ! READ POSTERIOR SAMPLES

  if (usrReadSamp) then
    call postrSampIO(chn,.NOT. OPT_OUTPUT,.FALSE.)
  else
    priAvail=.TRUE.
  end if


  ! PRINT STAR- AND DATA-RELATED INFORMATION

  if (.NOT. usrGenerateData) then
  
    if (usrReadSamp) then
      call printPostrSampProps(OPT_GENERAL) ! print properties of saved-samples file (general)
      if (dtProps%nSets == 0) call err("no data sets given or data sets different from original ones")
    end if

    if (dtProps%nSets > 0) call printDtMdlProps

    ! [R]JD
    if (dtProps%timesAreJD) then
      call printMsg(UNIT%cons+UNIT%log,"All further times refer to the Reduced Julian Date scale (RJD = JD - 2,400,000 d)")
    else
      call printMsg(UNIT%cons+UNIT%log,"All further times refer to the Julian Date scale")
    end if

  end if


  ! INITIALIZE RANDOM NUMBER GENERATOR

  call startSubsect(UNIT%cons+UNIT%log,"Initialising random number generator")
  if (keepSeed) call printMsg(UNIT%cons+UNIT%log,"Keeping old seed value")
  call iniRNG
  call printMsg(UNIT%cons+UNIT%log,"Done")
  call endSubsect(UNIT%cons+UNIT%log) ! Initialising random number generator



  ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
  !
  !   HANDLING THE DATA
  !
  ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·


  ! FOR PLOTTING

  if (.NOT. (onlySpdMeas .OR. usrGenerateData) .AND. filOutputPlt) then
    filNmGnupltCmdsNoDirs=trim(runIdent)//SUFFX_GNUPLT
    filNmGnupltCmds=trim(outDirCSV)//filNmGnupltCmdsNoDirs
    call tryOpenFilRW("Gnuplot",filNmGnupltCmds,UN_GNUPLT,"replace","write")
  end if


  ! GENERATE DATA

  if (usrGenerateData) then

    call startSubsect(UNIT%cons+UNIT%log,"Generating data")
    call generateData
    call printMsg(UNIT%cons+UNIT%log,"Done")
    call endSubsect(UNIT%cons+UNIT%log)

  else if (onlyPlotData) then

    call undef(psdNPlUndef)
    curFilNmPrefx=trim(runIdent)
    call startSubsect(UNIT%cons+UNIT%log,"Plotting data")
    call plotDataAndModel(setFlg([FLG%onlyDt,FLG%center]),tyCoord2DDbl(0._dp,0._dp))
    call endSubsect(UNIT%cons+UNIT%log)

  else


  ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
  !
  !   THE MAIN WORK
  !
  ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·


    ! PREPARE PARAMETERS FOR THE FIRST TIME

    call prepAnalysis


    ! GO THROUGH ALL THE #-PLANETS MODES

    hdr=mrg("Preparing speed measurement","Starting analysis",onlySpdMeas)

    do iNPlMd=1,nNPlMds

      if (mdlProps%bin) then
        call startSubsect(UNIT%cons+UNIT%log,trim(hdr)//" in binary mode")
      else
        call startSubsect(UNIT%cons+UNIT%log,trim(hdr)//" in mode "//trim(niceFmt(iNPlMd)))
        call printQty(UNIT%cons+UNIT%log,"# planets",nPl(iNPlMd))
      end if

      curNPl=nPl(iNPlMd)
      psdNPl=merge(1,curNPl,mdlProps%bin)
      psdNPlUndef=merge(VAL_UNDEF_INT,curNPl,mdlProps%bin)
      psdNPlReq=psdNPlUndef
      call undef(parEst(iNPlMd,:)%mode)

      if (usrReadSamp) then
        call postrSampIO(chn,.NOT.OPT_OUTPUT,.FALSE.,modeTmp,psdNPlReq) ! curNPars not yet set, thus the need for modeTmp
        if (usrPeriodogram .AND. .NOT. priAvail) call err("cannot calculate periodogram when prior is unknown")
        call printPostrSampProps(.NOT. OPT_GENERAL)
      end if

      call updtRuntm(iNPlMd)
      curFilNmPrefx=trim(runIdent)//trim(plStr(iNPlMd,psdNPlUndef,"_",prepend=.TRUE.))
      call prepNPlMd
      parEst(iNPlMd,1:curNPars)%mode=modeTmp(1:curNPars)
      
      call printPriRges
      
      if (usrReadSamp) then
        call calcLikh(chn)
      else
        call multiPT(chn,parEst(iNPlMd,1:curNPars)%mode,logPostrEv)
      end if

      if (.NOT. onlySpdMeas) then

        if (nPT > 1) call printPSR

        call chkPostrMode(chn)        ! also casts warning/error (error if in periodogram mode) if posterior mode has not been found

        if (usrSavSamp) call postrSampIO(chn,OPT_OUTPUT,iNPlMd == nNPlMds,parEst(iNPlMd,1:curNPars)%mode) ! save samples

        if (.NOT. usrNoInfc) then

          ! calc. summaries

          call calcSummaries(chn)      ! does not sort


          ! plots before sorting

          if (filOutputPlt) call makePostrPlots(.TRUE.,.TRUE.)


          ! calc. posterior probs. over hypercubes (needs chain unsorted)

          call calcProbAllPars(chn)


          ! calc. quantiles, marg. postrs., residuals and chi² (and sorts post-burn-in links)

          call calcQuantMaPoSort(chn)


          ! calc. posterior probs. over intervals (needs chain sorted)

          call calcProb(chn)


          ! find best posterior summary types for all pars. (for case where posterior mode has not been found)

          call findBestEst


          ! calc. chi²

          call calcChiSq


          ! make model predictions

          call calcAllPred(chn)


          ! calculate residuals, using best pars. summaries

          if (dtProps%nRec > 0 .AND. .NOT. usrPeriodogram) call calcResid


          ! posterior summaries

          call printSummaries


          ! plots after sorting

          if (filOutputPlt) call makePostrPlots(.FALSE.,.FALSE.)

        end if                         ! .NOT. usrNoInfc

      end if                           ! .NOT. onlySpdMeas

      call endSubsect(UNIT%cons+UNIT%log) ! analysis

    end do                             ! iNPlMd


    curFilNmPrefx=trim(runIdent)



    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·
    !
    !   FINAL SUMMARIES
    !
    ! · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · · ·


    ! general summaries
    
    if (.NOT. usrNoInfc) call printSummaries(.TRUE.)
    
  end if                             ! (usrGenerateData)

  
  ! plots

  if (.NOT. (usrGenerateData .OR. onlySpdMeas)) call doPlotting


  ! date & time

  call date_and_time(values=dateEndPrg)
  call printQty(UNIT%cons+UNIT%log,"Local date & time",fmtDate(dateEndPrg))

  call printQty(UNIT%cons+UNIT%log,"Total runtime [h:m:s]",fmtTime(int(omp_get_wtime()-tmStartPrg)))



  ! finish

  call endSubsect(UNIT%cons+UNIT%log,msg=["BASE finished.","BASE finished."])


  ! send email

  if (.NOT. usrGenerateData .AND. len_trim(mailAddrFin) > 0) call sendMailFin(hostName=hostName,runIdent=runIdent)


end program main
