module strings                                              ! Strings

! Routines:
! 
! undefValStr            Returns: value signifying an undefined variable of a certain type
! undefStr               Returns: value signifying an undefined variable of a certain type
! isUndefStr             Returns: whether argument is undefined
! isDefStr               Returns: whether a variable is defined
! setIfUndefStr          Sets variable to a given value if variable is undefined
! beginsWith             Returns: whether b is the first substring of a of length len(b)
! endsWith               Returns: whether b is the last substring of a of length len(b)
! lCase                  Returns: lower-case variant of a string
! lCaseFst               Returns: given string with first character converted to lower case
! uCase                  Returns: upper-case variant of a string
! uCaseFst               Returns: given string with first character converted to upper case
! nSubStr                Returns: number of (nonoverlapping) occurrences of b in a
! rmLead                 Returns: string str with leading substring subStr deleted if present
! rmTrail                Returns: string str with trailing substring subStr deleted if present
! rmTrailN               Returns: string str with n trailing characters deleted
! xtndStr                Returns: string extended by spaces to a given length
! allEqStr               Returns: whether all elements of an array are equal
! mrgStr                 Returns: one of two arguments depending on a boolean value
! idStr                  Returns: unique string according to an index
! isNumeric              Returns: whether a string represents a numeric value
! intFromStr             Returns: number represented by string
! realFromStr            Returns: number represented by string
! dblFromStr             Returns: number represented by string
! xplodeInt              Parses str for occurrences of sep and assigns the in-between values to arr
! xplodeReal             Parses str for occurrences of sep and assigns the in-between values to arr
! xplodeDbl              Parses str for occurrences of sep and assigns the in-between values to arr
! xplodeStr              Parses str for occurrences of sep and assigns the in-between substrings to arr
! fmtIntNr               Returns: number formatted according to options
! fmtRealNr              Returns: number formatted according to options
! fmtDblNr               Returns: number formatted according to options
! fmtQPNr                Returns: number formatted according to options
! fmtMatrixReal          Returns: matrix formatted according to options
! fmtMatrixDbl           Returns: matrix formatted according to options
! fmtMatrixStr           Returns: matrix formatted according to options
! getNiceIntFmt          Returns a format string and the total number of resulting digits for 'nr'
! getNiceInt2Fmt         Returns a format string and the total number of resulting digits for 'nr'
! getNiceInt8Fmt         Returns a format string and the total number of resulting digits for 'nr'
! getNiceRealFmt         Returns a format string and the total number of resulting digits for 'nr'
! getNiceDblFmt          Returns a format string and the total number of resulting digits for 'nr'
! getNiceQPFmt           Returns a format string and the total number of resulting digits for 'nr'
! niceInt                Returns: nicely formatted number
! niceInt2               Returns: nicely formatted number
! niceInt8               Returns: nicely formatted number
! niceReal               Returns: nicely formatted number
! niceDbl                Returns: nicely formatted number
! niceQP                 Returns: nicely formatted number

  use basicIO
  use basicTypes
  use etc
  use undefVals
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !=====================================================[ PARAMETERS/VARS. ]========================================================

  integer,parameter :: &
   STR_LEN_STD=128, &                ! used as length that should be sufficient for most "standard" strings
   STR_LEN_LONG=1000, &
   STR_LEN_LONGEST=200000, &
   MAX_LEN_PATH=STR_LEN_LONG
  character(*),parameter :: &
   STR_TRUE="True", &
   STR_FALSE="False", &
   CHARS_TRUE="ty"


  !=====================================================[ TYPE DEFINITIONS ]========================================================

  type tyCoord2DStr
    character(STR_LEN_STD) :: x,y
  end type


!––––––––––––––––––––––––––––––––––––––––––––––––––––––––––[ INTERFACES ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––


  ! UNDEFINED STRINGS

  interface undefVal
    module procedure undefValStr
  end interface

   interface undef
    module procedure undefStr
  end interface

  interface isUndef
    module procedure isUndefStr
  end interface

  interface isDef
    module procedure isDefStr
  end interface

  interface setIfUndef
    module procedure setIfUndefStr
  end interface


  ! BASIC PROCEDURES

  interface allEq
    module procedure allEqStr
  end interface

  interface mrg
    module procedure mrgStr
  end interface


  ! STRINGS AND NUMBERS

  interface xplode
    module procedure xplodeInt
    module procedure xplodeReal
    module procedure xplodeDbl
    module procedure xplodeStr
  end interface

  interface fmtNr
    module procedure fmtIntNr
    module procedure fmtRealNr
    module procedure fmtDblNr
    module procedure fmtQPNr
  end interface

  interface fmtMatrix
    module procedure fmtMatrixReal
    module procedure fmtMatrixDbl
    module procedure fmtMatrixStr
  end interface

  interface niceFmt
    module procedure niceInt
    module procedure niceInt2
    module procedure niceInt8
    module procedure niceReal
    module procedure niceDbl
    module procedure niceQP
  end interface


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


  !====================================================[ UNDEFINED STRINGS ]========================================================


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function undefValStr(var)
!
!   Returns:                         value signifying an undefined variable of a certain type
!
!   Arguments:
!     var                            variable of the relevant type
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character :: &
     undefValStr
    character(*) :: &
     var
    intent(in) :: &
     var

    undefValStr=VAL_UNDEF_CHR

  end function undefValStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefStr(var)
!
!   Returns:                         value signifying an undefined variable of a certain type
!
!   Arguments:
!     var                            variable of the relevant type
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*) :: &
     var
    intent(inout) :: &
     var

    var=repeat(VAL_UNDEF_CHR,len(var))

  end subroutine undefStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefStr(var)
!
!   Returns:                         whether argument is undefined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     i
    logical :: &
     isUndefStr
    character(*) :: &
     var
    intent(in) :: &
     var

    isUndefStr=.TRUE.
    do i=1,len(var)
      isUndefStr=isUndefStr .AND. var(i:i) == VAL_UNDEF_CHR
    end do

  end function isUndefStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefStr(var)
!
!   Returns:                         whether a variable is defined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isDefStr
    character(*) :: &
     var
    intent(in) :: &
     var

    isDefStr=.NOT.isUndef(var)

  end function isDefStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefStr(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var)) var=val

  end subroutine setIfUndefStr




  !====================================================[ BASIC PROCEDURES ]=========================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function beginsWith(a,b,pos,longer)
!
!   Returns:                         whether b is the first substring of a of length len(b)
!
!   Arguments:
!     a                              string
!     b                              string
!     pos                            position after b (may be undefined)
!     longer                         whether .FALSE. should be returned if a and b have the same length
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     lA, &
     lB, &
     pos
    logical :: &
     beginsWith, &
     longer, &
     tstL
    character(*) :: &
     a, &
     b
    optional :: &
     longer, &
     pos
    intent(in) :: &
     a, &
     b, &
     longer
    intent(out) :: &
     pos

    lA=len(a)
    lB=len(b)
    tstL=present(longer) .AND. longer

    if (lA < lB) then
      beginsWith=.FALSE.
    else
      beginsWith=a(1:lB) == b .AND. (.NOT. tstL .OR. (tstL .AND. lA > lB))
    end if
    if (present(pos)) then
      if (beginsWith) then
        pos=lB+1
      else
        call undef(pos)
      end if
    end if

  end function beginsWith




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function endsWith(a,b)
!
!   Returns:                         whether b is the last substring of a of length len(b)
!
!   Arguments:
!     a                              string
!     b                              string
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     lA, &
     lB
    logical :: &
     endsWith, &
     tstL
    character(*) :: &
     a, &
     b
    intent(in) :: &
     a, &
     b

    lA=len(a)
    lB=len(b)

    if (lA < lB) then
      endsWith=.FALSE.
    else
      endsWith=a(lA-lB+1:lA) == b
    end if

  end function endsWith




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function lCase(str)
!
!   Returns:                         lower-case variant of a string
!
!   Arguments:
!     str                            string to convert
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     iC, &
     i
    character(*) :: &
     str
    character(len(str)) :: &
     lCase
    intent(in) :: &
     str

    lCase=str
    do i=1,len(str)
      iC=ichar(str(i:i))
      if (iC >= 65 .AND. iC <= 90) lCase(i:i)=char(iC+32)
    end do

  end function lCase




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function lCaseFst(str)
!
!   Returns:                         given string with first character converted to lower case
!
!   Arguments:
!     str                            string to convert
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*) :: &
     str
    character(len(str)) :: &
     lCaseFst
    intent(in) :: &
     str

    lCaseFst=lCase(str(1:1))//str(2:)

  end function lCaseFst




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function uCase(str)
!
!   Returns:                         upper-case variant of a string
!
!   Arguments:
!     str                            string to convert
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     iC, &
     i
    character(*) :: &
     str
    character(len(str)) :: &
     uCase
    intent(in) :: &
     str

    uCase=str
    do i=1,len(str)
      iC=ichar(str(i:i))
      if (iC >= 97 .AND. iC <= 122) uCase(i:i)=char(iC-32)
    end do

  end function uCase




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function uCaseFst(str)
!
!   Returns:                         given string with first character converted to upper case
!
!   Arguments:
!     str                            string to convert
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*) :: &
     str
    character(len(str)) :: &
     uCaseFst
    intent(in) :: &
     str

    uCaseFst=uCase(str(1:1))//str(2:)

  end function uCaseFst




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function nSubStr(a,b)
!
!   Returns:                         number of (nonoverlapping) occurrences of b in a
!
!   Arguments:
!     a                              string
!     b                              string
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     nSubStr, &
     i, &
     lA, &
     lB
    character(*) :: &
     a, &
     b
    intent(in) :: &
     a, &
     b

    nSubStr=0
    lA=len(a)
    lB=len(B)
    if (lB <= lA) then
      i=1
      do
        if (a(i:i+lB-1) == b) then
          nSubStr=nSubStr+1
          i=i+lB
        else
          i=i+1
        end if
        if (i+lB-1 > lA) exit
      end do
    end if

  end function nSubStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function rmLead(str,subStr)
!
!   Returns:                         string str with leading substring subStr deleted if present
!
!   Arguments:
!     str                            string
!     subStr                         string
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*) :: &
     str, &
     subStr
    character(len(str)) :: &
     rmLead
    intent(in) :: &
     str, &
     subStr

    if (beginsWith(str,subStr)) then
      rmLead=str(len(subStr)+1:)
    else
      rmLead=str
    end if

  end function rmLead




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function rmTrail(str,subStr)
!
!   Returns:                         string str with trailing substring subStr deleted if present
!
!   Arguments:
!     str                            string
!     subStr                         string
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*) :: &
     str, &
     subStr
    character(len(str)) :: &
     rmTrail
    intent(in) :: &
     str, &
     subStr

    if (endsWith(str,subStr)) then
      rmTrail=str(1:len(str)-len(subStr))
    else
      rmTrail=str
    end if

  end function rmTrail




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function rmTrailN(str,n)
!
!   Returns:                         string str with n trailing characters deleted
!
!   Arguments:
!     str                            string
!     n                              natural number
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     n
    character(*) :: &
     str
    character(len(str)-n) :: &
     rmTrailN
    intent(in) :: &
     str, &
     n

    rmTrailN=str(1:len(str)-n)

  end function rmTrailN




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function xtndStr(str,l)
!
!   Returns:                         string extended by spaces to a given length
!
!   Arguments:
!     str                            string
!     l                              length
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     l
    character(*) :: &
     str
    character(max(len(str),l)) :: &
     xtndStr
    intent(in) :: &
     str, &
     l

    xtndStr=str

  end function xtndStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function allEqStr(arr)
!
!   Returns:                         whether all elements of an array are equal
!
!   Arguments:
!     arr                            array in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     allEqStr
    character(*) :: &
     arr(:)
    intent(in) :: &
     arr

    if (size(arr) <= 1) then
      allEqStr=.TRUE.
    else
      allEqStr=all(arr == arr(lbound(arr,1)))
    end if

  end function allEqStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function mrgStr(e1,e2,l)
!
!   Returns:                         one of two arguments depending on a boolean value
!
!   Arguments:
!     e1                             argument returned if (l)
!     e2                             argument returned if (.NOT. l)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     l
    character(*) :: &
     e1, &
     e2
    character(max(len(e1),len(e2))) :: &
     mrgStr
    intent(in) :: &
     l, &
     e1, &
     e2

    if (l) then
      mrgStr=e1
    else
      mrgStr=e2
    end if

  end function mrgStr




  !===================================================[ STRINGS AND NUMBERS ]=======================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function idStr(i)
!
!   Returns:                         unique string according to an index
!
!   Arguments:
!     i                              index
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     I_CHR_A=ichar("A"), &
     I_CHR_Z=ichar("Z"), &
     RGE=26
    integer :: &
     i, &
     n, &
     i1, &
     i2
    character(2) :: &
     idStr
    intent(in) :: &
     i


    idStr=""
    n=merge(2,1,I_CHR_A+i-1 > I_CHR_Z)
    if (i > RGE**2-1) then
      idStr="ZZ"
    else if (n == 1) then
      idStr=char(I_CHR_A+i-1)
    else
      i1=floor(1._dp*i/(floor(log(1._dp*i)/log(1._dp*RGE))*RGE))
      i2=mod(i,RGE)+1
      idStr=char(I_CHR_A+i1-1)//char(I_CHR_A+i2-1)
    end if


  end function idStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function isNumeric(str)
!
!   Returns:                         whether a string represents a numeric value
!
!   Arguments:
!     str                            string in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    character(*),parameter :: &
     STR_NUM="0123456789", &
     STR_NUM_DOT=STR_NUM//"."
    integer :: &
     p1, &
     p2
    logical :: &
     isNumeric
    character(*) :: &
     str
    character(len(str)) :: &
     adjStr
    intent(in) :: &
     str

    adjStr=adjustl(str)

    if (len_trim(adjStr) == 1) then
      isNumeric=scan(STR_NUM,adjStr(1:1)) > 0
    else if (len_trim(adjStr) > 1) then
      p1=scan(STR_NUM_DOT,adjStr(1:1))
      p2=scan(STR_NUM_DOT,adjStr(2:2))
      isNumeric=p1 > 0 .AND. p2 > 0 .AND. .NOT. beginsWith(str,"..")
    else
      isNumeric=.FALSE.
    end if

  end function isNumeric




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function intFromStr(str)
!
!   Returns:                         number represented by string
!
!   Arguments:
!     str                            string
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     intFromStr
    character(*) :: &
     str
    intent(in) :: &
     str

    read(str,*) intFromStr

  end function intFromStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function realFromStr(str)
!
!   Returns:                         number represented by string
!
!   Arguments:
!     str                            string
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     realFromStr
    character(*) :: &
     str
    intent(in) :: &
     str

    read(str,*) realFromStr

  end function realFromStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function dblFromStr(str)
!
!   Returns:                         number represented by string
!
!   Arguments:
!     str                            string
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     dblFromStr
    character(*) :: &
     str
    intent(in) :: &
     str

    read(str,*) dblFromStr

  end function dblFromStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine xplodeInt(str,sep,arr)
!
!   parses str for occurrences of sep and assigns the in-between values to arr
!   n: # values contained in str (if n < size(arr,1), not all values will be assigned to arr)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     arr(:), &
     l, &
     i, &
     lS, &
     n, &
     pos, &
     pos2, &
     allocStat
    real(dp) :: &
     tmp
    character(*) :: &
     str, &
     sep
    allocatable :: &
     arr
    intent(in) :: &
     str, &
     sep

    l=len(str)
    lS=len(sep)

    n=nSubStr(str,sep)+1
    if (allocated(arr)) deallocate(arr)
    allocate(arr(n),stat=allocStat)
    call testAllocErr(allocStat,67)

    pos=1
    do i=1,n
      pos2=pos-1+index(str(pos:),sep)
      if (pos2 == pos-1) pos2=l+1
      read(str(pos:pos2-1),*) tmp
      arr(i)=int(tmp)
      pos=pos2+lS
    end do

  end subroutine xplodeInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine xplodeReal(str,sep,arr)
!
!   parses str for occurrences of sep and assigns the in-between values to arr
!   n: # values contained in str (if n < size(arr,1), not all values will be assigned to arr)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     l, &
     i, &
     lS, &
     n, &
     pos, &
     pos2, &
     allocStat
    real :: &
     arr(:)
    real(dp) :: &
     tmp
    character(*) :: &
     str, &
     sep
    allocatable :: &
     arr
    intent(in) :: &
     str, &
     sep

    l=len(str)
    lS=len(sep)

    n=nSubStr(str,sep)+1
    if (allocated(arr)) deallocate(arr)
    allocate(arr(n),stat=allocStat)
    call testAllocErr(allocStat,68)

    pos=1
    do i=1,n
      pos2=pos-1+index(str(pos:),sep)
      if (pos2 == pos-1) pos2=l+1
      read(str(pos:pos2-1),*) tmp
      arr(i)=real(tmp,sp)
      pos=pos2+lS
    end do

  end subroutine xplodeReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine xplodeDbl(str,sep,arr)
!
!   parses str for occurrences of sep and assigns the in-between values to arr
!   n: # values contained in str (if n < size(arr,1), not all values will be assigned to arr)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     l, &
     i, &
     lS, &
     n, &
     pos, &
     pos2, &
     allocStat
    real(dp) :: &
     arr(:)
    character(*) :: &
     str, &
     sep
    allocatable :: &
     arr
    intent(in) :: &
     str, &
     sep

    l=len(str)
    lS=len(sep)

    n=nSubStr(str,sep)+1
    if (allocated(arr)) deallocate(arr)
    allocate(arr(n),stat=allocStat)
    call testAllocErr(allocStat,69)

    pos=1
    do i=1,n
      pos2=pos-1+index(str(pos:),sep)
      if (pos2 == pos-1) pos2=l+1
      read(str(pos:pos2-1),*) arr(i)
      pos=pos2+lS
    end do

  end subroutine xplodeDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine xplodeStr(str,sep,arr)
!
!   parses str for occurrences of sep and assigns the in-between substrings to arr
!   n: # substrings contained in str (if n < size(arr,1), not all substrings will be assigned to arr)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     l, &
     i, &
     lS, &
     n, &
     pos, &
     pos2, &
     allocStat
    character(STR_LEN_STD) :: &
     arr(:)
    character(*) :: &
     str, &
     sep
    allocatable :: &
     arr
    intent(in) :: &
     str, &
     sep

    l=len(str)
    lS=len(sep)

    n=nSubStr(str,sep)+1
    if (allocated(arr)) deallocate(arr)
    allocate(arr(n),stat=allocStat)
    call testAllocErr(allocStat,69)

    pos=1
    do i=1,n
      pos2=pos-1+index(str(pos:),sep)
      if (pos2 == pos-1) pos2=l+1
      arr(i)=str(pos:pos2-1)
      pos=pos2+lS
    end do

  end subroutine xplodeStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtIntNr(nr,nDigs,flags)
!
!   Returns:                         number formatted according to options
!
!   Arguments:
!     nur                            number to be formatted
!     nDigs                          total # output digits
!     flags                          option flags
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     nr, &
     nDigs, &
     oriNDigs, &
     flags
    logical :: &
     useAst
    character(nDigs) :: &
     fmtIntNr
    character(STR_LEN_LONG) :: &
     cTmp
    intent(in) :: &
     nr, &
     nDigs, &
     flags

    if (isUndef(nr)) then

      fmtIntNr=STR_UNDEF_BY_LEN(min(size(STR_UNDEF_BY_LEN),max(1,nDigs)))

    else

      if (nr < 0) then
        useAst=.TRUE.
      else
        if (nr == 0) then
          oriNDigs=1
        else
          oriNDigs=floor(log10(1.*nr))+1
        end if
        if (oriNDigs > nDigs) then
          useAst=.TRUE.
        else
          useAst=.FALSE.
        end if
      end if

      if (useAst) then
        fmtIntNr=repeat("*",nDigs)
      else
        write(cTmp,"(I20)") nr
        cTmp=adjustl(cTmp)
        if (getFlg(flags,FLG%use0)) then
          fmtIntNr=repeat("0",nDigs-oriNDigs)//trim(cTmp)
        else if (getFlg(flags,FLG%adjR)) then
          fmtIntNr=repeat(" ",nDigs-oriNDigs)//trim(cTmp)
        else
          fmtIntNr=trim(cTmp)
        end if
      end if

    end if

  end function fmtIntNr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtRealNr(nr,fmt,nDigs,flags)
!
!   Returns:                         number formatted according to options
!
!   Arguments:
!     nur                            number to be formatted
!     nDigs                          total # output digits
!     flags                          option flags
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     nDigs, &
     oriNDigs, &
     i, &
     nDigsUsd, &
     flags
    real :: &
     nr
    character(nDigs) :: &
     fmtRealNr
    character(STR_LEN_LONG) :: &
     cTmp
    character(*) :: &
     fmt
    intent(in) :: &
     nr, &
     fmt, &
     nDigs, &
     flags

    if (isUndef(nr)) then

      fmtRealNr=STR_UNDEF_BY_LEN(min(size(STR_UNDEF_BY_LEN),max(1,nDigs)))

    else

      write(cTmp,fmt) nr
      cTmp=adjustl(cTmp)
      oriNDigs=len_trim(cTmp)
      nDigsUsd=nDigs

      if (oriNDigs > nDigsUsd) then
        fmtRealNr=repeat("*",nDigsUsd)
      else
        if (nr < 0) then
          cTmp=cTmp(2:)
          nDigsUsd=nDigsUsd-1
          oriNDigs=oriNDigs-1
        end if

        if (getFlg(flags,FLG%rm0)) then
          do i=oriNDigs,2,-1
            if (cTmp(i:i) /= "0") exit
            cTmp=cTmp(:i-1)
          end do
        end if

        if (getFlg(flags,FLG%use0)) then
          fmtRealNr=trim(merge("-"," ",nr < 0))//repeat("0",nDigsUsd-oriNDigs)//trim(cTmp)
        else if (getFlg(flags,FLG%adjR)) then
          fmtRealNr=repeat(" ",nDigsUsd-oriNDigs)//trim(merge("-"," ",nr < 0))//trim(cTmp)
        else
          fmtRealNr=trim(merge("-"," ",nr < 0))//trim(cTmp)
        end if

        ! remove trailing dot
        fmtRealNr=rmTrail(trim(fmtRealNr),".")

      end if

    end if

  end function fmtRealNr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtDblNr(nr,fmt,nDigs,flags)
!
!   Returns:                         number formatted according to options
!
!   Arguments:
!     nur                            number to be formatted
!     nDigs                          total # output digits
!     flags                          option flags
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     nDigs, &
     oriNDigs, &
     i, &
     nDigsUsd, &
     flags
    real(dp) :: &
     nr
    character(nDigs) :: &
     fmtDblNr
    character(STR_LEN_LONG) :: &
     cTmp
    character(*) :: &
     fmt
    intent(in) :: &
     nr, &
     fmt, &
     nDigs, &
     flags

    if (isUndef(nr)) then

      fmtDblNr=STR_UNDEF_BY_LEN(min(size(STR_UNDEF_BY_LEN),max(1,nDigs)))

    else

      write(cTmp,fmt) nr
      cTmp=adjustl(cTmp)
      oriNDigs=len_trim(cTmp)
      nDigsUsd=nDigs

      if (oriNDigs > nDigsUsd) then
        fmtDblNr=repeat("*",nDigsUsd)
      else
        if (nr < 0) then
          cTmp=cTmp(2:)
          nDigsUsd=nDigsUsd-1
          oriNDigs=oriNDigs-1
        end if

        if (getFlg(flags,FLG%rm0)) then
          do i=oriNDigs,2,-1
            if (cTmp(i:i) /= "0") exit
            cTmp=cTmp(:i-1)
          end do
        end if

        if (getFlg(flags,FLG%use0)) then
          fmtDblNr=trim(merge("-"," ",nr < 0))//repeat("0",nDigsUsd-oriNDigs)//trim(cTmp)
        else if (getFlg(flags,FLG%adjR)) then
          fmtDblNr=repeat(" ",nDigsUsd-oriNDigs)//trim(merge("-"," ",nr < 0))//trim(cTmp)
        else
          fmtDblNr=trim(merge("-"," ",nr < 0))//trim(cTmp)
        end if

        ! remove trailing dot
        fmtDblNr=rmTrail(trim(fmtDblNr),".")

      end if

    end if

  end function fmtDblNr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtQPNr(nr,fmt,nDigs,flags)
!
!   Returns:                         number formatted according to options
!
!   Arguments:
!     nur                            number to be formatted
!     nDigs                          total # output digits
!     flags                          option flags
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     nDigs, &
     oriNDigs, &
     i, &
     nDigsUsd, &
     flags
    real(qp) :: &
     nr
    character(nDigs) :: &
     fmtQPNr
    character(STR_LEN_LONG) :: &
     cTmp
    character(*) :: &
     fmt
    intent(in) :: &
     nr, &
     fmt, &
     nDigs, &
     flags

    if (isUndef(nr)) then

      fmtQPNr=STR_UNDEF_BY_LEN(min(size(STR_UNDEF_BY_LEN),max(1,nDigs)))

    else

      write(cTmp,fmt) nr
      cTmp=adjustl(cTmp)
      oriNDigs=len_trim(cTmp)
      nDigsUsd=nDigs

      if (oriNDigs > nDigsUsd) then
        fmtQPNr=repeat("*",nDigsUsd)
      else
        if (nr < 0) then
          cTmp=cTmp(2:)
          nDigsUsd=nDigsUsd-1
          oriNDigs=oriNDigs-1
        end if

        if (getFlg(flags,FLG%rm0)) then
          do i=oriNDigs,2,-1
            if (cTmp(i:i) /= "0") exit
            cTmp=cTmp(:i-1)
          end do
        end if

        if (getFlg(flags,FLG%use0)) then
          fmtQPNr=trim(merge("-"," ",nr < 0))//repeat("0",nDigsUsd-oriNDigs)//trim(cTmp)
        else if (getFlg(flags,FLG%adjR)) then
          fmtQPNr=repeat(" ",nDigsUsd-oriNDigs)//trim(merge("-"," ",nr < 0))//trim(cTmp)
        else
          fmtQPNr=trim(merge("-"," ",nr < 0))//trim(cTmp)
        end if

        ! remove trailing dot
        fmtQPNr=rmTrail(trim(fmtQPNr),".")

      end if

    end if

  end function fmtQPNr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtMatrixReal(nRows,arr,hdr)
!
!   Returns:                         matrix formatted according to options
!
!   Arguments:
!     nRows                          # rows of matrix
!     arr                            matrix
!     hdr                            header to be included in result
!
!   Requirements:
!     - hdr, if present, must have the same size in the second dimension (i.e. # cols) as arr
!
!   arr, hdr ARE CONTIGUOUS (as of rev. 115)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     arr(:,:)
    integer :: &
     iC, &
     iR, &
     c, &
     r, &
     nRows
    character(*) :: &
     hdr(:,:)
    character(STR_LEN_LONG) :: &
     fmtMatrixReal(nRows), &
     tmpStr(size(arr,1),size(arr,2))
    optional :: &
     hdr
    intent(in) :: &
     arr, &
     nRows, &
     hdr

    r=size(arr,1)
    c=size(arr,2)

    do iR=1,r
      do iC=1,c
        tmpStr(iR,iC)=niceFmt(arr(iR,iC))
      end do
    end do
    
    if (present(hdr)) then
      fmtMatrixReal=fmtMatrix(nRows,tmpStr,hdr)
    else
      fmtMatrixReal=fmtMatrix(nRows,tmpStr)
    end if

  end function fmtMatrixReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtMatrixDbl(nRows,arr,hdr)
!
!   Returns:                         matrix formatted according to options
!
!   Arguments:
!     nRows                          # rows of matrix
!     arr                            matrix
!     hdr                            header to be included in result
!
!   Requirements:
!     - hdr, if present, must have the same size in the second dimension (i.e. # cols) as arr
!
!   arr, hdr ARE CONTIGUOUS (as of rev. 115)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     arr(:,:)
    integer :: &
     iC, &
     iR, &
     c, &
     r, &
     nRows
    character(*) :: &
     hdr(:,:)
    character(STR_LEN_LONG) :: &
     fmtMatrixDbl(nRows), &
     tmpStr(size(arr,1),size(arr,2))
    optional :: &
     hdr
    intent(in) :: &
     arr, &
     nRows, &
     hdr

    r=size(arr,1)
    c=size(arr,2)

    do iR=1,r
      do iC=1,c
        tmpStr(iR,iC)=niceFmt(arr(iR,iC))
      end do
    end do
    
    if (present(hdr)) then
      fmtMatrixDbl=fmtMatrix(nRows,tmpStr,hdr)
    else
      fmtMatrixDbl=fmtMatrix(nRows,tmpStr)
    end if

  end function fmtMatrixDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtMatrixStr(nRows,arr,hdr)
!
!   Returns:                         matrix formatted according to options
!
!   Arguments:
!     nRows                          # rows of matrix
!     arr                            matrix
!     hdr                            header to be included in result
!
!   Requirements:
!     - hdr, if present, must have the same size in the second dimension (i.e. # cols) as arr
!
!   arr, hdr ARE CONTIGUOUS (as of rev. 115)
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     MIN_SEP=1
    character,parameter :: &
     SEP=" "
    character(*) :: &
     arr(:,:), &
     hdr(:,:)
    integer :: &
     nRows, &
     iC, &
     iR, &
     c, &
     r, &
     maxLen(size(arr,2)), &
     rH, &
     nxtPad, &
     allocStat
    character(STR_LEN_LONG) :: &
     fmtMatrixStr(nRows), &
     tmpStrH(:,:)
    allocatable :: &
     tmpStrH
    optional :: &
     hdr
    intent(in) :: &
     arr, &
     nRows, &
     hdr

    r=size(arr,1)
    c=size(arr,2)
    rH=merge(size(hdr,1),0,present(hdr))

    do iC=1,c
      maxLen(iC)=max(maxval(len_trim(arr(:,iC))),merge(maxval(len_trim(hdr(:,iC))),-HUGE_INT,present(hdr)))
    end do

    fmtMatrixStr=""

    if (present(hdr)) then
      allocate(tmpStrH(size(hdr,1),size(hdr,2)),stat=allocStat)
      call testAllocErr(allocStat,56)
      tmpStrH=hdr
      do iR=1,rH
        do iC=1,c
          if (iC == 1) then
            fmtMatrixStr(iR)=tmpStrH(iR,iC)
          else
            fmtMatrixStr(iR)=trim(fmtMatrixStr(iR))//repeat(SEP,nxtPad)//tmpStrH(iR,iC)
          end if
          nxtPad=MIN_SEP+maxLen(iC)-len_trim(tmpStrH(iR,iC))
        end do
      end do
      deallocate(tmpStrH)
    end if

    do iR=rH+1,rH+r
      do iC=1,c
        if (iC == 1) then
          fmtMatrixStr(iR)=arr(iR-rH,iC)
        else
          fmtMatrixStr(iR)=trim(fmtMatrixStr(iR))//repeat(SEP,nxtPad)//arr(iR-rH,iC)
        end if
        nxtPad=MIN_SEP+maxLen(iC)-len_trim(arr(iR-rH,iC))
      end do
    end do

  end function fmtMatrixStr




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine getNiceIntFmt(nr,fmt,nDigs,minNDigs,noExp)
!
!   returns a format string and the total number of resulting digits for 'nr'
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     EXP_OVRHD=6
    integer :: &
     nr, &
     minNDigs, &
     nDigs, &
     nSgn, &
     nDecDigs, &
     nDAbs
    logical :: &
     noExp
    character(3) :: &
     cNDigs, &
     cNDecDigs
    character(*) :: &
     fmt
    optional :: &
     minNDigs, &
     noExp                           ! min. # counting digits (relevant for exponential format)
    intent(in) :: &
     nr, &
     minNDigs, &
     noExp
    intent(out) :: &
     fmt, &
     nDigs


    ! # chars for sign
    if (nr < 0) then
      nSgn=1
    else
      nSgn=0
    end if

    nDAbs=floor(log10(abs(1.*nr)))+1

    ! exponential or integer format?

    if (.NOT. (present(noExp) .AND. noExp) .AND. (present(minNDigs) .AND. nDAbs > minNDigs+EXP_OVRHD)) then ! exponential format shorter
      nDecDigs=minNDigs-1
      nDigs=nSgn+1+nDecDigs+EXP_OVRHD
      write(cNDigs,"(I3)") nDigs
      write(cNDecDigs,"(I3)") nDecDigs
      fmt="(ES"//cNDigs//"."//cNDecDigs//")"
    else
      if (nr == 0) then
        nDigs=1
      else
        nDigs=nSgn+nDAbs
      end if
      write(cNDigs,"(I3)") nDigs
      fmt="(I"//cNDigs//")"
    end if

  end subroutine getNiceIntFmt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine getNiceInt2Fmt(nr,fmt,nDigs,minNDigs,noExp)
!
!   returns a format string and the total number of resulting digits for 'nr'
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     EXP_OVRHD=6
    integer :: &
     minNDigs, &
     nDigs, &
     nSgn, &
     nDecDigs, &
     nDAbs
    integer(i2b) :: &
     nr
    logical :: &
     noExp
    character(3) :: &
     cNDigs, &
     cNDecDigs
    character(*) :: &
     fmt
    optional :: &
     minNDigs, &
     noExp                           ! min. # counting digits (relevant for exponential format)
    intent(in) :: &
     nr, &
     minNDigs, &
     noExp
    intent(out) :: &
     fmt, &
     nDigs


    ! # chars for sign
    if (nr < 0) then
      nSgn=1
    else
      nSgn=0
    end if

    nDAbs=floor(log10(abs(1.*nr)))+1

    ! exponential or integer format?

    if (.NOT. (present(noExp) .AND. noExp) .AND. (present(minNDigs) .AND. nDAbs > minNDigs+EXP_OVRHD)) then ! exponential format shorter
      nDecDigs=minNDigs-1
      nDigs=nSgn+1+nDecDigs+EXP_OVRHD
      write(cNDigs,"(I3)") nDigs
      write(cNDecDigs,"(I3)") nDecDigs
      fmt="(ES"//cNDigs//"."//cNDecDigs//")"
    else
      if (nr == 0) then
        nDigs=1
      else
        nDigs=nSgn+nDAbs
      end if
      write(cNDigs,"(I3)") nDigs
      fmt="(I"//cNDigs//")"
    end if

  end subroutine getNiceInt2Fmt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine getNiceInt8Fmt(nr,fmt,nDigs,minNDigs,noExp)
!
!   returns a format string and the total number of resulting digits for 'nr'
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     EXP_OVRHD=6
    integer :: &
     minNDigs, &
     nDigs, &
     nSgn, &
     nDecDigs, &
     nDAbs
    integer(i8b) :: &
     nr
    logical :: &
     noExp
    character(3) :: &
     cNDigs, &
     cNDecDigs
    character(*) :: &
     fmt
    optional :: &
     minNDigs, &
     noExp                           ! min. # counting digits (relevant for exponential format)
    intent(in) :: &
     nr, &
     minNDigs, &
     noExp
    intent(out) :: &
     fmt, &
     nDigs


    ! # chars for sign
    if (nr < 0) then
      nSgn=1
    else
      nSgn=0
    end if

    nDAbs=floor(log10(abs(1.*nr)))+1

    ! exponential or integer format?

    if (.NOT. (present(noExp) .AND. noExp) .AND. (present(minNDigs) .AND. nDAbs > minNDigs+EXP_OVRHD)) then ! exponential format shorter
      nDecDigs=minNDigs-1
      nDigs=nSgn+1+nDecDigs+EXP_OVRHD
      write(cNDigs,"(I3)") nDigs
      write(cNDecDigs,"(I3)") nDecDigs
      fmt="(ES"//cNDigs//"."//cNDecDigs//")"
    else
      if (nr == 0) then
        nDigs=1
      else
        nDigs=nSgn+nDAbs
      end if
      write(cNDigs,"(I3)") nDigs
      fmt="(I"//cNDigs//")"
    end if

  end subroutine getNiceInt8Fmt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine getNiceRealFmt(nr,fmt,nDigs,remove0,noExp)
!
!   returns a format string and the total number of resulting digits for 'nr'
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     EXP_OVRHD=2, &
     F_OVRHD=1, &
     N_DOT=1
    integer :: &
     nDigs, &
     nDecDigs, &
     nZeros, &
     nSgn, &
     i, &
     nDReal, &
     nDAbs, &
     nDExp, &
     lstDecDigPos
    real :: &
     nr
    logical :: &
     remove0, &
     noExp
    character(2) :: &
     cFmtSpecExp
    character(3) :: &
     cFmtSpec, &
     cNDigs, &
     cNDecDigs
    character(*) :: &
     fmt
    character(STR_LEN_STD) :: &
     cTest
    optional :: &
     remove0, &
     noExp
    intent(in) :: &
     nr, &
     remove0, &
     noExp
    intent(out) :: &
     fmt, &
     nDigs


    ! # chars for sign
    if (nr < 0) then
      nSgn=1
    else
      nSgn=0
    end if

    nZeros=-floor(log10(abs(nr)))    ! # leading zeros for nr < 1 (including the 0 left of decimal point)
    nDAbs=floor(log10(abs(nr)))+1
    nDReal=precision(nr)+1
    nDExp=merge(floor(log10(1.*nZeros))+1,floor(log10(nDAbs-1.))+1,nZeros > 0)

    ! exponential or floating-point format?

    if (.NOT. (present(noExp) .AND. noExp) .AND. (nDAbs > nDReal .OR. nZeros > N_ZEROS_FLOAT_MAX) .AND. &
     nr /= 0) then
      ! floating-point format would be longer than # counting digits; or # leading zeros would exceed maximum
      nDecDigs=nDReal-1
      nDigs=nSgn+nDReal+N_DOT+EXP_OVRHD+nDExp
      lstDecDigPos=nDigs-(EXP_OVRHD+nDExp)
      cFmtSpec="ES"
      write(cFmtSpecExp,"('E',I1)") nDExp
    else
      if (nr == 0.) then
        nDecDigs=0
        nDigs=1+F_OVRHD
      else if (abs(nr) < 1.) then
        nDecDigs=nZeros-1+nDReal
        nDigs=nSgn+1+nDecDigs+F_OVRHD
      else
        nDecDigs=max(0,nDReal-nDAbs)
        nDigs=nSgn+max(nDReal,nDAbs)+F_OVRHD
      end if
      lstDecDigPos=nDigs
      cFmtSpec="F"
      cFmtSpecExp=""
    end if

    ! format string

    write(cNDigs,"(I3)") nDigs
    write(cNDecDigs,"(I3)") nDecDigs
    fmt="("//trim(cFmtSpec)//trim(adjustl(cNDigs))//"."//trim(adjustl(cNDecDigs))//trim(cFmtSpecExp)//")"

    ! remove any trailing zeros after the decimal point

    if (nDecDigs > 0 .AND. .NOT. (present(remove0) .AND. .NOT. remove0)) then
      write(cTest,fmt) nr
      do i=lstDecDigPos,nSgn+N_DOT+2,-1
        if (cTest(i:i) /= "0") exit
      end do
      if (i < lstDecDigPos) then
        nDecDigs=nDecDigs-(lstDecDigPos-i)
        nDigs=nDigs-(lstDecDigPos-i)
        write(cNDigs,"(I3)") nDigs
        write(cNDecDigs,"(I3)") nDecDigs
        fmt="("//trim(cFmtSpec)//trim(adjustl(cNDigs))//"."//trim(adjustl(cNDecDigs))//trim(cFmtSpecExp)//")"
      end if
    end if

  end subroutine getNiceRealFmt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine getNiceDblFmt(nr,fmt,nDigs,remove0,noExp)
!
!   returns a format string and the total number of resulting digits for 'nr'
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     EXP_OVRHD=2, &
     F_OVRHD=1, &
     N_DOT=1
    integer :: &
     nDigs, &
     nDecDigs, &
     nZeros, &
     nSgn, &
     i, &
     nDDbl, &
     nDAbs, &
     nDExp, &
     lstDecDigPos
    real(dp) :: &
     nr
    logical :: &
     remove0, &
     noExp
    character(2) :: &
     cFmtSpecExp
    character(3) :: &
     cFmtSpec, &
     cNDigs, &
     cNDecDigs
    character(*) :: &
     fmt
    character(STR_LEN_STD) :: &
     cTest
    optional :: &
     remove0, &
     noExp
    intent(in) :: &
     nr, &
     remove0, &
     noExp
    intent(out) :: &
     fmt, &
     nDigs


    ! # chars for sign
    if (nr < 0) then
      nSgn=1
    else
      nSgn=0
    end if

    nZeros=-floor(log10(abs(nr)))    ! # leading zeros for nr < 1 (including the 0 left of decimal point)
    nDAbs=floor(log10(abs(nr)))+1
    nDDbl=precision(nr)+1
    nDExp=merge(floor(log10(1.*nZeros))+1,floor(log10(nDAbs-1.))+1,nZeros > 0)

    ! exponential or floating-point format?

    if (.NOT. (present(noExp) .AND. noExp) .AND. (nDAbs > nDDbl .OR. nZeros > N_ZEROS_FLOAT_MAX) .AND. &
     nr /= 0) then
      ! floating-point format would be longer than # counting digits; or # leading zeros would exceed maximum
      nDecDigs=nDDbl-1
      nDigs=nSgn+nDDbl+N_DOT+EXP_OVRHD+nDExp
      lstDecDigPos=nDigs-(EXP_OVRHD+nDExp)
      cFmtSpec="ES"
      write(cFmtSpecExp,"('E',I1)") nDExp
    else
      if (nr == 0.) then
        nDecDigs=0
        nDigs=1+F_OVRHD
      else if (abs(nr) < 1.) then
        nDecDigs=nZeros-1+nDDbl
        nDigs=nSgn+1+nDecDigs+F_OVRHD
      else
        nDecDigs=max(0,nDDbl-nDAbs)
        nDigs=nSgn+max(nDDbl,nDAbs)+F_OVRHD
      end if
      lstDecDigPos=nDigs
      cFmtSpec="F"
      cFmtSpecExp=""
    end if

    ! format string

    write(cNDigs,"(I3)") nDigs
    write(cNDecDigs,"(I3)") nDecDigs
    fmt="("//trim(cFmtSpec)//trim(adjustl(cNDigs))//"."//trim(adjustl(cNDecDigs))//trim(cFmtSpecExp)//")"

    ! remove any trailing zeros after the decimal point

    if (nDecDigs > 0 .AND. .NOT. (present(remove0) .AND. .NOT. remove0)) then
      write(cTest,fmt) nr
      do i=lstDecDigPos,nSgn+N_DOT+2,-1
        if (cTest(i:i) /= "0") exit
      end do
      if (i < lstDecDigPos) then
        nDecDigs=nDecDigs-(lstDecDigPos-i)
        nDigs=nDigs-(lstDecDigPos-i)
        write(cNDigs,"(I3)") nDigs
        write(cNDecDigs,"(I3)") nDecDigs
        fmt="("//trim(cFmtSpec)//trim(adjustl(cNDigs))//"."//trim(adjustl(cNDecDigs))//trim(cFmtSpecExp)//")"
      end if
    end if

  end subroutine getNiceDblFmt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine getNiceQPFmt(nr,fmt,nDigs,remove0,noExp)
!
!   returns a format string and the total number of resulting digits for 'nr'
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer,parameter :: &
     EXP_OVRHD=2, &
     F_OVRHD=1, &
     N_DOT=1
    integer :: &
     nDigs, &
     nDecDigs, &
     nZeros, &
     nSgn, &
     i, &
     nDQP, &
     nDAbs, &
     nDExp, &
     lstDecDigPos
    real(qp) :: &
     nr
    logical :: &
     remove0, &
     noExp
    character(2) :: &
     cFmtSpecExp
    character(3) :: &
     cFmtSpec, &
     cNDigs, &
     cNDecDigs
    character(*) :: &
     fmt
    character(STR_LEN_STD) :: &
     cTest
    optional :: &
     remove0, &
     noExp
    intent(in) :: &
     nr, &
     remove0, &
     noExp
    intent(out) :: &
     fmt, &
     nDigs


    ! # chars for sign
    if (nr < 0) then
      nSgn=1
    else
      nSgn=0
    end if

    nZeros=-floor(log10(abs(nr)))    ! # leading zeros for nr < 1 (including the 0 left of decimal point)
    nDAbs=floor(log10(abs(nr)))+1
    nDQP=precision(nr)+1
    nDExp=merge(floor(log10(1.*nZeros))+1,floor(log10(nDAbs-1.))+1,nZeros > 0)

    ! exponential or floating-point format?

    if (.NOT. (present(noExp) .AND. noExp) .AND. (nDAbs > nDQP .OR. nZeros > N_ZEROS_FLOAT_MAX) .AND. &
     nr /= 0) then
      ! floating-point format would be longer than # counting digits; or # leading zeros would exceed maximum
      nDecDigs=nDQP-1
      nDigs=nSgn+nDQP+N_DOT+EXP_OVRHD+nDExp
      lstDecDigPos=nDigs-(EXP_OVRHD+nDExp)
      cFmtSpec="ES"
      write(cFmtSpecExp,"('E',I1)") nDExp
    else
      if (nr == 0.) then
        nDecDigs=0
        nDigs=1+F_OVRHD
      else if (abs(nr) < 1.) then
        nDecDigs=nZeros-1+nDQP
        nDigs=nSgn+1+nDecDigs+F_OVRHD
      else
        nDecDigs=max(0,nDQP-nDAbs)
        nDigs=nSgn+max(nDQP,nDAbs)+F_OVRHD
      end if
      lstDecDigPos=nDigs
      cFmtSpec="F"
      cFmtSpecExp=""
    end if

    ! format string

    write(cNDigs,"(I3)") nDigs
    write(cNDecDigs,"(I3)") nDecDigs
    fmt="("//trim(cFmtSpec)//trim(adjustl(cNDigs))//"."//trim(adjustl(cNDecDigs))//trim(cFmtSpecExp)//")"

    ! remove any trailing zeros after the decimal point

    if (nDecDigs > 0 .AND. .NOT. (present(remove0) .AND. .NOT. remove0)) then
      write(cTest,fmt) nr
      do i=lstDecDigPos,nSgn+N_DOT+2,-1
        if (cTest(i:i) /= "0") exit
      end do
      if (i < lstDecDigPos) then
        nDecDigs=nDecDigs-(lstDecDigPos-i)
        nDigs=nDigs-(lstDecDigPos-i)
        write(cNDigs,"(I3)") nDigs
        write(cNDecDigs,"(I3)") nDecDigs
        fmt="("//trim(cFmtSpec)//trim(adjustl(cNDigs))//"."//trim(adjustl(cNDecDigs))//trim(cFmtSpecExp)//")"
      end if
    end if

  end subroutine getNiceQPFmt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function niceInt(nr,minNDigs,lenArg,noExp)
!
!   Returns:                         nicely formatted number
!
!   Arguments:
!     nr                             number to be formatted
!     minNDigs                       min. # output digits
!     lenArg                         output length used in case of undefined number
!     noExp                          whether exponential representation is unwanted
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    integer :: &
     nDigs, &
     minNDigs, &
     lenArg, &
     nr
    logical :: &
     noExp
    character(15) :: &
     fmt
    character(STR_LEN_LONG) :: &
     niceInt
    optional :: &
     lenArg, &
     minNDigs, &
     noExp
    intent(in) :: &
     nr, &
     minNDigs, &
     lenArg, &
     noExp

    if (isUndef(nr)) then
      niceInt=STR_UNDEF_BY_LEN(min(merge(max(1,lenArg),huge(1),present(lenArg)),size(STR_UNDEF_BY_LEN)))
    else
      if (present(minNDigs)) then
        call getNiceIntFmt(nr,fmt,nDigs,minNDigs,noExp=(present(noExp) .AND. noExp)) ! allows exponential format
      else
        call getNiceIntFmt(nr,fmt,nDigs,noExp=present(noExp) .AND. noExp)
      end if
      if (fmt(2:2) == "I") then
        write(niceInt,fmt) nr
      else
        write(niceInt,fmt) 1.*nr
      end if
    end if

  end function niceInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function niceInt2(nr,minNDigs,lenArg,noExp)
!
!   Returns:                         nicely formatted number
!
!   Arguments:
!     nr                             number to be formatted
!     minNDigs                       min. # output digits
!     lenArg                         output length used in case of undefined number
!     noExp                          whether exponential representation is unwanted
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    integer :: &
     nDigs, &
     minNDigs, &
     lenArg
    integer(i2b) :: &
     nr
    logical :: &
     noExp
    character(15) :: &
     fmt
    character(STR_LEN_LONG) :: &
     niceInt2
    optional :: &
     lenArg, &
     minNDigs, &
     noExp
    intent(in) :: &
     nr, &
     minNDigs, &
     lenArg, &
     noExp

    if (isUndef(nr)) then
      niceInt2=STR_UNDEF_BY_LEN(min(merge(max(1,lenArg),huge(1),present(lenArg)),size(STR_UNDEF_BY_LEN)))
    else
      if (present(minNDigs)) then
        call getNiceInt2Fmt(nr,fmt,nDigs,minNDigs,noExp=(present(noExp) .AND. noExp)) ! allows exponential format
      else
        call getNiceInt2Fmt(nr,fmt,nDigs,noExp=present(noExp) .AND. noExp)
      end if
      if (fmt(2:2) == "I") then
        write(niceInt2,fmt) nr
      else
        write(niceInt2,fmt) 1.*nr
      end if
    end if

  end function niceInt2




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function niceInt8(nr,minNDigs,lenArg,noExp)
!
!   Returns:                         nicely formatted number
!
!   Arguments:
!     nr                             number to be formatted
!     minNDigs                       min. # output digits
!     lenArg                         output length used in case of undefined number
!     noExp                          whether exponential representation is unwanted
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    integer :: &
     nDigs, &
     minNDigs, &
     lenArg
    integer(i8b) :: &
     nr
    logical :: &
     noExp
    character(15) :: &
     fmt
    character(STR_LEN_LONG) :: &
     niceInt8
    optional :: &
     lenArg, &
     minNDigs, &
     noExp
    intent(in) :: &
     nr, &
     minNDigs, &
     lenArg, &
     noExp

    if (isUndef(nr)) then
      niceInt8=STR_UNDEF_BY_LEN(min(merge(max(1,lenArg),huge(1),present(lenArg)),size(STR_UNDEF_BY_LEN)))
    else
      if (present(minNDigs)) then
        call getNiceInt8Fmt(nr,fmt,nDigs,minNDigs,noExp=(present(noExp) .AND. noExp)) ! allows exponential format
      else
        call getNiceInt8Fmt(nr,fmt,nDigs,noExp=present(noExp) .AND. noExp)
      end if
      if (fmt(2:2) == "I") then
        write(niceInt8,fmt) nr
      else
        write(niceInt8,fmt) 1.*nr
      end if
    end if

  end function niceInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function niceReal(nr,remove0,lenArg,noExp)
!
!   Returns:                         nicely formatted number
!
!   Arguments:
!     nr                             number to be formatted
!     minNDigs                       min. # output digits
!     lenArg                         output length used in case of undefined number
!     noExp                          whether exponential representation is unwanted
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    integer :: &
     nDigs, &
     lenArg
    real :: &
     nr
    logical :: &
     remove0, &
     noExp
    character(15) :: &
     fmt
    character(STR_LEN_LONG) :: &
     niceReal
    optional :: &
     lenArg, &
     remove0, &
     noExp
    intent(in) :: &
     nr, &
     remove0, &
     lenArg, &
     noExp

    if (isUndef(nr)) then
      niceReal=STR_UNDEF_BY_LEN(min(merge(max(1,lenArg),huge(1),present(lenArg)),size(STR_UNDEF_BY_LEN)))
    else
      call getNiceRealFmt(nr,fmt,nDigs,remove0=.NOT.present(remove0) .OR. (present(remove0) .AND. remove0), &
       noExp=present(noExp) .AND. noExp)
      niceReal=fmtNr(nr,fmt,nDigs,NO_FLAGS)
    end if

  end function niceReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function niceDbl(nr,remove0,lenArg,noExp)
!
!   Returns:                         nicely formatted number
!
!   Arguments:
!     nr                             number to be formatted
!     minNDigs                       min. # output digits
!     lenArg                         output length used in case of undefined number
!     noExp                          whether exponential representation is unwanted
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    integer :: &
     nDigs, &
     lenArg
    real(dp) :: &
     nr
    logical :: &
     remove0, &
     noExp
    character(15) :: &
     fmt
    character(STR_LEN_LONG) :: &
     niceDbl
    optional :: &
     lenArg, &
     remove0, &
     noExp
    intent(in) :: &
     nr, &
     remove0, &
     lenArg, &
     noExp

    if (isUndef(nr)) then
      niceDbl=STR_UNDEF_BY_LEN(min(merge(max(1,lenArg),huge(1),present(lenArg)),size(STR_UNDEF_BY_LEN)))
    else
      call getNiceDblFmt(nr,fmt,nDigs,remove0=.NOT.present(remove0) .OR. (present(remove0) .AND. remove0), &
       noExp=present(noExp) .AND. noExp)
      niceDbl=fmtNr(nr,fmt,nDigs,NO_FLAGS)
    end if

  end function niceDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function niceQP(nr,remove0,lenArg,noExp)
!
!   Returns:                         nicely formatted number
!
!   Arguments:
!     nr                             number to be formatted
!     minNDigs                       min. # output digits
!     lenArg                         output length used in case of undefined number
!     noExp                          whether exponential representation is unwanted
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    integer :: &
     nDigs, &
     lenArg
    real(qp) :: &
     nr
    logical :: &
     remove0, &
     noExp
    character(15) :: &
     fmt
    character(STR_LEN_LONG) :: &
     niceQP
    optional :: &
     lenArg, &
     remove0, &
     noExp
    intent(in) :: &
     nr, &
     remove0, &
     lenArg, &
     noExp

    if (isUndef(nr)) then
      niceQP=STR_UNDEF_BY_LEN(min(merge(max(1,lenArg),huge(1),present(lenArg)),size(STR_UNDEF_BY_LEN)))
    else
      call getNiceQPFmt(nr,fmt,nDigs,remove0=.NOT.present(remove0) .OR. (present(remove0) .AND. remove0), &
       noExp=present(noExp) .AND. noExp)
      niceQP=fmtNr(nr,fmt,nDigs,NO_FLAGS)
    end if

  end function niceQP




end module
