module time                                              ! Date and time

! Routines:
! 
! dateandtimeFromTime8   Returns: date and time in format of 'values' argument of date_and_time intrinsic
! fmtDate                Returns: formatted date, time, and time-zone ofset from UTC
! fmtTime                Returns: time in seconds formatted as hours, minutes, and seconds

  use basicTypes
  use omp_lib
  use etc
  use strings
  use intervals
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !========================================================[ CONSTANTS ]============================================================

  integer,parameter :: &
   N_SPD_MEAS=3

  real,parameter :: &
   SPD_SUBMEAS_DUR_S_MIN=2., &       ! min. duration of a sub-measurement of speed
   SPD_MEAS_RUNTM_RATIO=1./100       ! ratio (duration of speed measurement):(remaining runtime)

  real(dp),parameter :: &
   UTH_MULTIPT_EST_SPD_MIN=50._dp

  type(tyIntvlReal),parameter :: &
   SPD_MEAS_DUR_S_RGE=tyIntvlReal(5.,10.)   ! min./max. duration of speed measurement


  !========================================================[ VARIABLES ]============================================================

  integer :: &
   dateStartPrg(8), &
   dateEndPrg(8)

  real(dp) :: &
   curSpdEst, &
   tmStartPrg


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function dateandtimeFromTime8(tm8)
!
!   Returns:                         date and time in format of 'values' argument of date_and_time intrinsic
!
!   Arguments:
!     tm8                            time in format of 'time' argument of ltime intrinsic, i.e. as returned by time8 intrinsic
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     tm8, &
     dateandtimeFromTime8(8), &
     tmLtime(9)
    intent(in) :: &
     tm8

    call ltime(tm8,tmLtime)

    dateandtimeFromTime8(1)=1900+tmLtime(6)
    dateandtimeFromTime8(2)=1+tmLtime(5)
    dateandtimeFromTime8(3)=tmLtime(4)
    dateandtimeFromTime8(4)=dateStartPrg(4)
    dateandtimeFromTime8(5)=tmLtime(3)
    dateandtimeFromTime8(6)=tmLtime(2)
    dateandtimeFromTime8(7)=min(59,tmLtime(1))
    dateandtimeFromTime8(8)=0

  end function dateandtimeFromTime8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtDate(date,filNm,short)
!
!   Returns:                         formatted date, time, and time-zone ofset from UTC
!
!   Arguments:
!     date                           date and time as returned by dateandtimeFromTime8
!     filNm                          whether result should be suitable as part of a file name
!     short                          whether short output is requested
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     date(8), &
     tzMin, &
     tzHr
    logical :: &
     filNm, &
     short
    character :: &
     tZSgn
    character(29) :: &
     fmtDate
    optional :: &
     filNm, &
     short
    intent(in) :: &
     date, &
     filNm, &
     short

    if (present(filNm) .AND. filNm) then
      fmtDate=fmtNr(date(1),4,NO_FLAGS)//"-"//fmtNr(date(2),2,setFlg(FLG%use0))//"-"//fmtNr(date(3),2,setFlg(FLG%use0))//"_"// &
       trim(adjustl(fmtTime(date(5)*3600+date(6)*60+date(7),OPT_USE_DOTS,OPT_USE_0)))
    else if (present(short) .AND. short) then
      fmtDate=fmtNr(date(1),4,NO_FLAGS)//"-"//fmtNr(date(2),2,setFlg(FLG%use0))//"-"//fmtNr(date(3),2,setFlg(FLG%use0))//" "// &
       trim(adjustl(fmtTime(date(5)*3600+date(6)*60+date(7),use0=.TRUE.)))
    else
      if (date(4) < 0) then
        tZSgn="-"
        tZHr=floor(-date(4)/60.)
        tZMin=mod(-date(4),60)
      else
        tZSgn="+"
        tZHr=floor(date(4)/60.)
        tZMin=mod(date(4),60)
      end if
      fmtDate=fmtNr(date(1),4,NO_FLAGS)//"-"//fmtNr(date(2),2,setFlg(FLG%use0))//"-"//fmtNr(date(3),2,setFlg(FLG%use0))//" "// &
       trim(adjustl(fmtTime(date(5)*3600+date(6)*60+date(7))))// &
       " UTC"//tZSgn//trim(fmtNr(tZHr,2,NO_FLAGS))//trim(mrg(":"//fmtNr(tZMin,2,setFlg(FLG%use0)),"",tZMin > 0))
    end if

  end function fmtDate




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    function fmtTime(sArg,useDots,use0,hrs3)
!
!   Returns:                         time in seconds formatted as hours, minutes, and seconds
!
!   Arguments:
!     sArg                           # seconds corresponding to the time to be formatted
!     useDots                        whether dots should be used as separators
!     use0                           whether the result should be zero-padded
!     hrs3                           whether the hours field should have three digits
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     sArg, &
     s
    logical :: &
     useDots, &
     use0, &
     hrs3, &
     useZero, &
     h3
    character :: &
     sep
    character(10) :: &
     fmtTime
    optional :: &
     useDots, &
     use0, &
     hrs3
    intent(in) :: &
     sArg, &
     useDots, &
     use0, &
     hrs3


    if (isUndef(sArg)) then

      fmtTime=STR_UNDEF_BY_LEN(min(10,size(STR_UNDEF_BY_LEN)))

    else

      s=max(0,sArg)

      if (present(useDots) .AND. useDots) then
        sep="."
      else
        sep=":"
      end if

      useZero=present(use0) .AND. use0
      h3=present(hrs3) .AND. hrs3

      fmtTime=fmtNr(floor(s/3600.),merge(3,2,h3),setFlg([FLG%use0,FLG%adjR],[useZero,.TRUE.]))//sep// &
       fmtNr(floor(mod(s,3600)/60.),2,setFlg(FLG%use0))//sep//fmtNr(mod(s,60),2,setFlg(FLG%use0))

      end if

  end function fmtTime




end module
