module uncert                            ! Representing quantities with their uncertainties

! Routines:
! 
! undefUncEllDbl         Sets argument as undefined
! undefUncQtyQPDbl       Sets argument as undefined
! undefUncQtyDbl         Sets argument as undefined
! undefUncQtyQP          Sets argument as undefined
! isUndefUncQtyDbl       Returns: whether argument is undefined
! isUndefUncQtyQP        Returns: whether argument is undefined
! isDefUncQtyDbl         Returns: whether argument is defined
! isDefUncQtyQP          Returns: whether argument is defined

  use basicTypes
  use undefVals
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !========================================================[ CONSTANTS ]============================================================

  character(*),parameter :: &
   DESC_UNC_QTY_VAL="value", &
   DESC_UNC_QTY_UNC="uncertainty"


  !=====================================================[ TYPE DEFINITIONS ]========================================================
  
  type tyUncQtyDbl
    real(dp) :: val,unc
  end type

  type tyUncQtyQP
    real(qp) :: val,unc
  end type

  type tyUncEllDbl
    real(dp) :: a,recprASq,b,recprBSq,phi,sinPhi,cosPhi,sinSqPhi,cosSqPhi
    ! semi-major and -minor axes and their reciprocal values, and rotation angle of major axis of uncertainty ellipse,
    ! border of ellipse being defined by set of points where chi²-chi²_min=1 (cf. [07.04.2010/2])
  end type


!––––––––––––––––––––––––––––––––––––––––––––––––––––––––––[ INTERFACES ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  interface isDef
    module procedure isDefUncQtyDbl
    module procedure isDefUncQtyQP
  end interface

  interface isUndef
    module procedure isUndefUncQtyDbl
    module procedure isUndefUncQtyQP
  end interface

  interface undef
    module procedure undefUncEllDbl
    module procedure undefUncQtyDbl
    module procedure undefUncQtyQP
  end interface


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefUncEllDbl(coord)
!
!   sets argument as undefined
!
!   Arguments:
!     coord                          var to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyUncEllDbl) :: &
     coord
    intent(inout) :: &
     coord

    call undef(coord%a)
    call undef(coord%recprASq)
    call undef(coord%b)
    call undef(coord%recprBSq)
    call undef(coord%phi)
    call undef(coord%sinPhi)
    call undef(coord%cosPhi)
    call undef(coord%sinSqPhi)
    call undef(coord%cosSqPhi)

  end subroutine undefUncEllDbl



!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefUncQtyQPDbl(qty)
!
!   sets argument as undefined
!
!   Arguments:
!     qty                            var to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyUncQtyDbl) :: &
     qty
    intent(inout) :: &
     qty

    call undef(qty%val)
    call undef(qty%unc)

  end subroutine undefUncQtyQPDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefUncQtyDbl(qty)
!
!   sets argument as undefined
!
!   Arguments:
!     qty                            var to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyUncQtyDbl) :: &
     qty
    intent(inout) :: &
     qty

    call undef(qty%val)
    call undef(qty%unc)

  end subroutine undefUncQtyDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefUncQtyQP(qty)
!
!   sets argument as undefined
!
!   Arguments:
!     qty                            var to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    type(tyUncQtyQP) :: &
     qty
    intent(inout) :: &
     qty

    call undef(qty%val)
    call undef(qty%unc)

  end subroutine undefUncQtyQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefUncQtyDbl(qty)
!
!   Returns:                         whether argument is undefined
!
!   Arguments:
!     qty                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isUndefUncQtyDbl
    type(tyUncQtyDbl) :: &
     qty
    intent(in) :: &
     qty

    isUndefUncQtyDbl=isUndef(qty%val) ! uncertainty may be undefined independently

  end function isUndefUncQtyDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefUncQtyQP(qty)
!
!   Returns:                         whether argument is undefined
!
!   Arguments:
!     qty                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isUndefUncQtyQP
    type(tyUncQtyQP) :: &
     qty
    intent(in) :: &
     qty

    isUndefUncQtyQP=isUndef(qty%val) ! uncertainty may be undefined independently

  end function isUndefUncQtyQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefUncQtyDbl(qty)
!
!   Returns:                         whether argument is defined
!
!   Arguments:
!     qty                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isDefUncQtyDbl
    type(tyUncQtyDbl) :: &
     qty
    intent(in) :: &
     qty

    isDefUncQtyDbl=.NOT.isUndef(qty)

  end function isDefUncQtyDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefUncQtyQP(qty)
!
!   Returns:                         whether argument is defined
!
!   Arguments:
!     qty                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    logical :: &
     isDefUncQtyQP
    type(tyUncQtyQP) :: &
     qty
    intent(in) :: &
     qty

    isDefUncQtyQP=.NOT.isUndef(qty)

  end function isDefUncQtyQP




end module
