module undefVals                                 ! Undefined values of variables

! Routines:
! 
! undefValInt            Returns: value signifying an undefined variable of a certain type
! undefValInt1           Returns: value signifying an undefined variable of a certain type
! undefValInt2           Returns: value signifying an undefined variable of a certain type
! undefValInt8           Returns: value signifying an undefined variable of a certain type
! undefValReal           Returns: value signifying an undefined variable of a certain type
! undefValDbl            Returns: value signifying an undefined variable of a certain type
! undefValQP             Returns: value signifying an undefined variable of a certain type
! undefInt               Sets argument as undefined
! undefInt1              Sets argument as undefined
! undefInt2              Sets argument as undefined
! undefInt8              Sets argument as undefined
! undefReal              Sets argument as undefined
! undefDbl               Sets argument as undefined
! undefQP                Sets argument as undefined
! isUndefInt             Returns: whether argument is undefined
! isUndefInt1            Returns: whether argument is undefined
! isUndefInt2            Returns: whether argument is undefined
! isUndefInt8            Returns: whether argument is undefined
! isUndefReal            Returns: whether argument is undefined
! isUndefDbl             Returns: whether argument is undefined
! isUndefQP              Returns: whether argument is undefined
! isDefInt               Returns: whether argument is defined
! isDefInt1              Returns: whether argument is defined
! isDefInt2              Returns: whether argument is defined
! isDefInt8              Returns: whether argument is defined
! isDefReal              Returns: whether argument is defined
! isDefDbl               Returns: whether argument is defined
! isDefQP                Returns: whether argument is defined
! setIfUndefInt          Sets variable to a given value if variable is undefined
! setIfUndefInt1         Sets variable to a given value if variable is undefined
! setIfUndefInt2         Sets variable to a given value if variable is undefined
! setIfUndefInt8         Sets variable to a given value if variable is undefined
! setIfUndefReal         Sets variable to a given value if variable is undefined
! setIfUndefDbl          Sets variable to a given value if variable is undefined
! setIfUndefQP           Sets variable to a given value if variable is undefined
! setIfDefInt            Sets variable to a given value if variable is defined
! setIfDefInt1           Sets variable to a given value if variable is defined
! setIfDefInt2           Sets variable to a given value if variable is defined
! setIfDefInt8           Sets variable to a given value if variable is defined
! setIfDefReal           Sets variable to a given value if variable is defined
! setIfDefDbl            Sets variable to a given value if variable is defined
! setIfDefQP             Sets variable to a given value if variable is defined

  use basicTypes
  implicit none

!–––––––––––––––––––––––––––––––––––––––––––––––––[ DECLARATIONS AND DEFINITIONS ]––––––––––––––––––––––––––––––––––––––––––––––––––


  !========================================================[ CONSTANTS ]============================================================

  ! CAUTION: VAL_UNDEF_… should not be changed unless absolutely necessary, since this may invalidate old saved-samples files
  
  integer,parameter :: &
   VAL_UNDEF_INT=transfer(z'80000000',1)                ! -2147483648 = -huge(1)-1 (!)

  integer(i1b),parameter :: &
   VAL_UNDEF_INT1=transfer(z'80',1_i1b)                 ! -128 = -huge(1_i1b)-1 (!)

  integer(i2b),parameter :: &
   VAL_UNDEF_INT2=transfer(z'8000',1_i2b)               ! -32768 = -huge(1_i2b)-1 (!)

  integer(i8b),parameter :: &
   VAL_UNDEF_INT8=transfer(z'8000000000000000',1_i8b)   ! -9223372036854775808 = -huge(1_i8b)-1 (!)

  real(sp),parameter :: &
   VAL_UNDEF_REAL=transfer(z'FF7FFFFE',1.)              ! >~ -huge(1.)

  real(dp),parameter :: &
   VAL_UNDEF_DBL=transfer(z'FFEFFFFFFFFFFFFE',1._dp)    ! >~ -huge(1._dp)

  real(qp),parameter :: &
   VAL_UNDEF_QP=transfer(z'FFFEFFFFFFFFFFFFFFFE',1._qp) ! >~ -huge(1._qp)

  character,parameter :: &
   VAL_UNDEF_CHR=transfer(z'7f',"a")                    ! (nonprintable)

  character(*),parameter :: &
   STR_UNDEF="<undefined>", &
   STR_UNDEF_BY_LEN(11)=[character(11) :: "?","?","?","?","?","?","<undef>","<undef>","<undef>","<undef>","<undefined>"], &
   STR_UNDEF_FIL_NM="[undefined]", &
   STR_NONE="<none>"


!––––––––––––––––––––––––––––––––––––––––––––––––––––––––––[ INTERFACES ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  interface undefVal
    module procedure undefValReal
    module procedure undefValDbl
    module procedure undefValQP
    module procedure undefValInt
    module procedure undefValInt1
    module procedure undefValInt2
    module procedure undefValInt8
  end interface

  interface undef
    module procedure undefReal
    module procedure undefDbl
    module procedure undefQP
    module procedure undefInt
    module procedure undefInt1
    module procedure undefInt2
    module procedure undefInt8
  end interface

  interface isUndef
    module procedure isUndefReal
    module procedure isUndefDbl
    module procedure isUndefQP
    module procedure isUndefInt
    module procedure isUndefInt1
    module procedure isUndefInt2
    module procedure isUndefInt8
  end interface

  interface isDef
    module procedure isDefReal
    module procedure isDefDbl
    module procedure isDefQP
    module procedure isDefInt
    module procedure isDefInt1
    module procedure isDefInt2
    module procedure isDefInt8
  end interface

  interface setIfUndef
    module procedure setIfUndefInt
    module procedure setIfUndefInt1
    module procedure setIfUndefInt2
    module procedure setIfUndefInt8
    module procedure setIfUndefReal
    module procedure setIfUndefDbl
    module procedure setIfUndefQP
  end interface

  interface setIfDef
    module procedure setIfDefInt
    module procedure setIfDefInt1
    module procedure setIfDefInt2
    module procedure setIfDefInt8
    module procedure setIfDefReal
    module procedure setIfDefDbl
    module procedure setIfDefQP
  end interface


!–––––––––––––––––––––––––––––––––––––––––––––––––––––––[ SUBPROGRAM PART ]–––––––––––––––––––––––––––––––––––––––––––––––––––––––––

  contains


  !====================================================[ UNDEFINED VALUES ]=========================================================


!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function undefValInt(var)
!
!   Returns:                         value signifying an undefined variable of a certain type
!
!   Arguments:
!     var                            variable of the relevant type
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     var, &
     undefValInt
    intent(in) :: &
     var

    undefValInt=VAL_UNDEF_INT

  end function undefValInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function undefValInt1(var)
!
!   Returns:                         value signifying an undefined variable of a certain type
!
!   Arguments:
!     var                            variable of the relevant type
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i1b) :: &
     var, &
     undefValInt1
    intent(in) :: &
     var

    undefValInt1=VAL_UNDEF_INT1

  end function undefValInt1




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function undefValInt2(var)
!
!   Returns:                         value signifying an undefined variable of a certain type
!
!   Arguments:
!     var                            variable of the relevant type
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i2b) :: &
     var, &
     undefValInt2
    intent(in) :: &
     var

    undefValInt2=VAL_UNDEF_INT2

  end function undefValInt2




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function undefValInt8(var)
!
!   Returns:                         value signifying an undefined variable of a certain type
!
!   Arguments:
!     var                            variable of the relevant type
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     var, &
     undefValInt8
    intent(in) :: &
     var

    undefValInt8=VAL_UNDEF_INT8

  end function undefValInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function undefValReal(var)
!
!   Returns:                         value signifying an undefined variable of a certain type
!
!   Arguments:
!     var                            variable of the relevant type
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     var, &
     undefValReal
    intent(in) :: &
     var

    undefValReal=VAL_UNDEF_REAL

  end function undefValReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function undefValDbl(var)
!
!   Returns:                         value signifying an undefined variable of a certain type
!
!   Arguments:
!     var                            variable of the relevant type
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     var, &
     undefValDbl
    intent(in) :: &
     var

    undefValDbl=VAL_UNDEF_DBL

  end function undefValDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function undefValQP(var)
!
!   Returns:                         value signifying an undefined variable of a certain type
!
!   Arguments:
!     var                            variable of the relevant type
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(qp) :: &
     var, &
     undefValQP
    intent(in) :: &
     var

    undefValQP=VAL_UNDEF_QP

  end function undefValQP




  !==============================================[ SETTING VARIABLES UNDEFINED ]====================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefInt(var)
!
!   sets argument as undefined
!
!   Arguments:
!     var                            var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     var
    intent(inout) :: &
     var

    var=VAL_UNDEF_INT

  end subroutine undefInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefInt1(var)
!
!   sets argument as undefined
!
!   Arguments:
!     var                            var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i1b) :: &
     var
    intent(inout) :: &
     var

    var=VAL_UNDEF_INT1

  end subroutine undefInt1




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefInt2(var)
!
!   sets argument as undefined
!
!   Arguments:
!     var                            var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i2b) :: &
     var
    intent(inout) :: &
     var

    var=VAL_UNDEF_INT2

  end subroutine undefInt2




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefInt8(var)
!
!   sets argument as undefined
!
!   Arguments:
!     var                            var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     var
    intent(inout) :: &
     var

    var=VAL_UNDEF_INT8

  end subroutine undefInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefReal(var)
!
!   sets argument as undefined
!
!   Arguments:
!     var                            var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     var
    intent(inout) :: &
     var

    var=VAL_UNDEF_REAL

  end subroutine undefReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefDbl(var)
!
!   sets argument as undefined
!
!   Arguments:
!     var                            var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     var
    intent(inout) :: &
     var

    var=VAL_UNDEF_DBL

  end subroutine undefDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental subroutine undefQP(var)
!
!   sets argument as undefined
!
!   Arguments:
!     var                            var. to be set as undefined
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(qp) :: &
     var
    intent(inout) :: &
     var

    var=VAL_UNDEF_QP

  end subroutine undefQP




  !====================================================[ QUERYING STATUS ]==========================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefInt(var)
!
!   Returns:                         whether argument is undefined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     var
    logical :: &
     isUndefInt
    intent(in) :: &
     var

    isUndefInt=var == VAL_UNDEF_INT

  end function isUndefInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefInt1(var)
!
!   Returns:                         whether argument is undefined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i1b) :: &
     var
    logical :: &
     isUndefInt1
    intent(in) :: &
     var

    isUndefInt1=var == VAL_UNDEF_INT1

  end function isUndefInt1




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefInt2(var)
!
!   Returns:                         whether argument is undefined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i2b) :: &
     var
    logical :: &
     isUndefInt2
    intent(in) :: &
     var

    isUndefInt2=var == VAL_UNDEF_INT2

  end function isUndefInt2




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefInt8(var)
!
!   Returns:                         whether argument is undefined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     var
    logical :: &
     isUndefInt8
    intent(in) :: &
     var

    isUndefInt8=var == VAL_UNDEF_INT8

  end function isUndefInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefReal(var)
!
!   Returns:                         whether argument is undefined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     var
    logical :: &
     isUndefReal
    intent(in) :: &
     var

    isUndefReal=var == VAL_UNDEF_REAL

  end function isUndefReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefDbl(var)
!
!   Returns:                         whether argument is undefined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     var
    logical :: &
     isUndefDbl
    intent(in) :: &
     var

    isUndefDbl=var == VAL_UNDEF_DBL

  end function isUndefDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isUndefQP(var)
!
!   Returns:                         whether argument is undefined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(qp) :: &
     var
    logical :: &
     isUndefQP
    intent(in) :: &
     var

    isUndefQP=var == VAL_UNDEF_QP

  end function isUndefQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefInt(var)
!
!   Returns:                         whether argument is defined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     var
    logical :: &
     isDefInt
    intent(in) :: &
     var

    isDefInt=.NOT.isUndef(var)

  end function isDefInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefInt1(var)
!
!   Returns:                         whether argument is defined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i1b) :: &
     var
    logical :: &
     isDefInt1
    intent(in) :: &
     var

    isDefInt1=.NOT.isUndef(var)

  end function isDefInt1




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefInt2(var)
!
!   Returns:                         whether argument is defined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i2b) :: &
     var
    logical :: &
     isDefInt2
    intent(in) :: &
     var

    isDefInt2=.NOT.isUndef(var)

  end function isDefInt2




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefInt8(var)
!
!   Returns:                         whether argument is defined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     var
    logical :: &
     isDefInt8
    intent(in) :: &
     var

    isDefInt8=.NOT.isUndef(var)

  end function isDefInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefReal(var)
!
!   Returns:                         whether argument is defined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     var
    logical :: &
     isDefReal
    intent(in) :: &
     var

    isDefReal=.NOT.isUndef(var)

  end function isDefReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefDbl(var)
!
!   Returns:                         whether argument is defined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     var
    logical :: &
     isDefDbl
    intent(in) :: &
     var

    isDefDbl=.NOT.isUndef(var)

  end function isDefDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    elemental function isDefQP(var)
!
!   Returns:                         whether argument is defined
!
!   Arguments:
!     var                            variable in question
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(qp) :: &
     var
    logical :: &
     isDefQP
    intent(in) :: &
     var

    isDefQP=.NOT.isUndef(var)

  end function isDefQP




  !=================================================[ CONDITIONAL ASSIGNMENTS ]=====================================================




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefInt(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var)) var=val

  end subroutine setIfUndefInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefInt1(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i1b) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var)) var=val

  end subroutine setIfUndefInt1




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefInt2(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i2b) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var)) var=val

  end subroutine setIfUndefInt2




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefInt8(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var)) var=val

  end subroutine setIfUndefInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefReal(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var)) var=val

  end subroutine setIfUndefReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefDbl(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var)) var=val

  end subroutine setIfUndefDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfUndefQP(var,val)
!
!   sets variable to a given value if variable is undefined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(qp) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isUndef(var)) var=val

  end subroutine setIfUndefQP




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfDefInt(var,val)
!
!   sets variable to a given value if variable is defined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isDef(var)) var=val

  end subroutine setIfDefInt




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfDefInt1(var,val)
!
!   sets variable to a given value if variable is defined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i1b) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isDef(var)) var=val

  end subroutine setIfDefInt1




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfDefInt2(var,val)
!
!   sets variable to a given value if variable is defined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i2b) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isDef(var)) var=val

  end subroutine setIfDefInt2




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfDefInt8(var,val)
!
!   sets variable to a given value if variable is defined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    integer(i8b) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isDef(var)) var=val

  end subroutine setIfDefInt8




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfDefReal(var,val)
!
!   sets variable to a given value if variable is defined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isDef(var)) var=val

  end subroutine setIfDefReal




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfDefDbl(var,val)
!
!   sets variable to a given value if variable is defined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(dp) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isDef(var)) var=val

  end subroutine setIfDefDbl




!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::
!
!
    subroutine setIfDefQP(var,val)
!
!   sets variable to a given value if variable is defined
!
!   Arguments:
!     var                            var. in question
!     val                            value to be assigned
!
!
!:::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::::

    implicit none
    real(qp) :: &
     var, &
     val
    intent(inout) :: &
     var
    intent(in) :: &
     val

    if (isDef(var)) var=val

  end subroutine setIfDefQP




end module
